#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2020 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


from collections import namedtuple

from construct import CString, Struct
from construct import Byte, Bytes, this
from construct import Computed, Switch, Tell
from construct import Int32ul
from construct import Int32ub

from enum import IntEnum

from objutils.dwarf.encoding import ULEB


AttributeDescription = namedtuple("AttributeDescription", "tag value parameterType conv")


class Reader(IntEnum):
    """
    """
    INT32   = 1
    CSTRING = 2
    ULEB    = 3

Ident = {}

Tag_CPU_arch = {
    0:  "Pre-v4",
    1:  "ARM v4",
    2:  "ARM v4T",
    3:  "ARM v5T",
    4:  "ARM v5TE",
    5:  "ARM v5TEJ",
    6:  "ARM v6",
    7:  "ARM v6KZ",
    8:  "ARM v6T2",
    9:  "ARM v6K",
    10: "ARM v7",
    11: "ARM v6-M",
    12: "ARM v6S-M",
    13: "ARM v7E-M",
    14: "ARM v8",
}

Tag_CPU_arch_profile = {
    0x00: "Architecture profile is not applicable",
    0x41: "The application profile",
    0x52: "The real-time profile",
    0x4D: "The microcontroller profile",
    0x53: "Application or real-time profile",
}

Tag_ARM_ISA_use =  {
    0: "The user did not permit this entity to use ARM instructions",
    1: "The user intended that this entity could use ARM instructions",
}

Tag_THUMB_ISA_use = {
    0: "The user did not permit this entity to use Thumb instructions",
    1: "The user permitted this entity to use 16-bit Thumb instructions (including BL)",
    2: "32-bit Thumb instructions were permitted",
}

Tag_FP_arch = {
    0: "The user did not permit this entity to use instructions requiring FP hardware ",
    1: "The user permitted use of instructions from v1 of the floating point (FP) ISA ",
    2: "Use of the v2 FP ISA was permitted (implies use of the v1 FP ISA) ",
    3: "Use of the v3 FP ISA was permitted (implies use of the v2 FP ISA) ",
    4: "Use of the v3 FP ISA was permitted, but only citing registers D0-D15, S0-S31 ",
    5: "Use of the v4 FP ISA was permitted (implies use of the non-vector v3 FP ISA) ",
    6: "Use of the v4 FP ISA was permitted, but only citing registers D0-D15, S0-S31 ",
    7: "Use of the ARM v8-A FP ISA was permitted ",
    8: "Use of the ARM v8-A FP ISA was permitted, but only citing registers D0-D15, S0-S31",
}

Tag_WMMX_arch = {
    0: "The user did not permit this entity to use WMMX",
    1: "The user permitted this entity to use WMMX v1",
    2: "The user permitted this entity to use WMMX v2",
}

Tag_Advanced_SIMD_arch = {
    0: "The user did not permit this entity to use the Advanced SIMD Architecture (Neon)",
    1: "Use of the Advanced SIMDv1 Architecture (Neon) was permitted",
    2: "Use of Advanced SIMDv2 Architecture (Neon) (with half-precision floating-point and fused MAC operations) was permitted",
    3: "Use of the ARM v8-A Advanced SIMD Architecture (Neon) was permitted",
    4: "Use of the ARM v8.1-A Advanced SIMD Architecture (Neon) was permitted",
}

Tag_FP_HP_extension = {
    0: "The user intended half-precision floating point instructions may be used if they exist in the available FP and ASIMD instruction sets as indicated by Tag_FP_arch and Tag_ASIMD_arch ",
    1: "Use of the optional half-precision extension to VFPv3/Advanced SIMDv1 was permitted",
}

Tag_CPU_unaligned_access = {
    0: "The user did not intend this entity to make unaligned data accesses",
    1: "The user intended that this entity might make v6-style unaligned data accesses",
}

Tag_T2EE_use = {
    0: "No use of T2EE extension was permitted, or no information is available",
    1: "Use of the T2EE extension was permitted",
}

Tag_Virtualization_use = {
    0: "No use of any virtualization extension was permitted, or no information available",
    1: "Use of the TrustZone extension was permitted",
    2: "Use of the virtualization extensions (HVC, ERET) were permitted",
    3: "Use of TrustZone and virtualization extensions were permitted",
}

Tag_MPextension_use = {
    0: "No use of ARM v7 MP extension was permitted, or no information available.",
    1: "Use of the ARM v7 MP extension was permitted.",
}

Tag_DIV_use = {
    0: "The user intended divide instructions may be used if they exist, or no explicit information recorded. This code was permitted to use SDIV and UDIV if the instructions are guaranteed present in the architecture, as indicated by Tag_CPU_arch and Tag_CPU_arch_profile.",
    1: "This code was explicitly not permitted to use SDIV or UDIV.",
    2: "This code was permitted to use SDIV and UDIV in the ARM and Thumb ISAs; the instructions are present as an optional architectural extension above the base architecture implied by Tag_CPU_arch and Tag_CPU_arch_profile.",
}

Tag_PCS_config = {
    0: "No standard configuration used, or no information recorded ",
    1: "Bare platform configuration ",
    2: "Linux application configuration ",
    3: "Linux DSO configuration ",
    4: "Palm OS 2004 configuration ",
    5: "Reserved to future Palm OS configuration ",
    6: "Symbian OS 2004 configuration ",
    7: "Reserved to future Symbian OS configuration",
}

Tag_ABI_PCS_R9_use = {
    0: "R9 used as V6 (just another callee-saved register, implied by omitting the tag) ",
    1: "R9 used as SB, a global Static Base register",
    2: "R9 used as a Thread Local Storage (TLS) pointer",
    3: "R9 not used at all by code associated with the attributed entity",
}

Tag_ABI_PCS_RW_data = {
    0: "RW static data was permitted to be addressed absolutely",
    1: "RW static data was only permitted to be addressed PC-relative",
    2: "RW static data was only permitted to be addressed SB-relative",
    3: "The user did not permit this entity to use RW static data",
}

Tag_ABI_PCS_RO_data = {
    0: "RO static data was permitted to be addressed absolutely",
    1: "RO static data was only permitted to be addressed PC-relative",
    2: "The user did not permit this entity to use RO static data",
}

Tag_ABI_PCS_GOT_use = {
    0: "The user did not permit this entity to import static data",
    1: "The user permitted this entity to address imported data directly",
    2: "The user permitted this entity to address imported data indirectly (e.g. via a GOT)",
}

Tag_ABI_PCS_wchar_t = {
    0: "The user prohibited the use of wchar_t when building this entity",
    2: "The user intended the size of wchar_t to be 2",
    4: "The user intended the size of wchar_t to be 4",
}

Tag_ABI_enum_size = {
    0: "The user prohibited the use of enums when building this entity ",
    1: "Enum values occupy the smallest container big enough to hold all their values ",
    2: "The user intended Enum containers to be 32-bit ",
    3: "The user intended that every enumeration visible across an ABI-complying interface contains a value needing 32 bits to encode it; other enums can be containerized",
}

Tag_ABI_align_needed = {
    0: "The user did not permit code to depend the alignment of 8-byte data or data with extended (> 8-byte) alignment",
    1: "Code was permitted to depend on the 8-byte alignment of 8-byte data items",
    2: "Code was permitted to depend on the 4-byte alignment of 8-byte data items",
    3: "Reserved",
    #n (in 4..12) Code was permitted to depend on the 8-byte alignment of 8-byte data items and the alignment of data items having up to 2n-byte extended alignment
}

Tag_ABI_align_preserved = {
    0: "The user did not require code to preserve 8-byte alignment of 8-byte data objects",
    1: "Code was required to preserve 8-byte alignment of 8-byte data objects",
    2: "Code was required to preserve 8-byte alignment of 8-byte data objects and to ensure (SP MOD 8) = 0 at all instruction boundaries (not just at function calls)",
    3: "Reserved",
    #n (in 4..12) Code was required to preserve the alignments of case 2 and the alignment of data items having up to 2n-byte extended alignment.
}

Tag_ABI_FP_rounding = {
    0: "The user intended this code to use the IEEE 754 round to nearest rounding mode",
    1: "The user permitted this code to choose the IEEE 754 rounding mode at run time",
}

Tag_ABI_FP_denormal = {
    0: "The user built this code knowing that denormal numbers might be flushed to (+) zero",
    1: "The user permitted this code to depend on IEEE 754 denormal numbers",
    2: "The user permitted this code to depend on the sign of a flushed-to-zero number being preserved in the sign of 0",
}

Tag_ABI_FP_exceptions = {
    0: "The user intended that this code should not check for inexact results",
    1: "The user permitted this code to check the IEEE 754 inexact exception",
}

Tag_ABI_FP_user_exceptions = {
    0: "The user intended that this code should not enable or use IEEE user exceptions",
    1: "The user permitted this code to enables and use IEEE 754 user exceptions",
}

Tag_ABI_FP_number_model = {
    0: "The user intended that this code should not use floating point numbers ",
    1: "The user permitted this code to use IEEE 754 format normal numbers only ",
    2: "The user permitted numbers, infinities, and one quiet NaN (see [RTABI]) ",
    3: "The user permitted this code to use all the IEEE 754-defined FP encodings",
}

Tag_ABI_FP_16bit_format = {
    0: "The user intended that this entity should not use 16-bit floating point numbers ",
    1: "Use of IEEE 754 (draft, November 2006) format 16-bit FP numbers was permitted ",
    2: "Use of VFPv3/Advanced SIMD alternative format 16-bit FP numbers was permitted",
}

Tag_ABI_HardFP_use = {
    0: "The user intended that FP use should be implied by Tag_FP_arch",
    1: "The user intended this code to execute on the single-precision variant derived from Tag_FP_arch",
    2: "Reserved",
    3: "The user intended that FP use should be implied by Tag_FP_arch (Note: This is a deprecated duplicate of the default encoded by 0)",
}

Tag_ABI_VFP_args = {
    0: "The user intended FP parameter/result passing to conform to AAPCS, base variant",
    1: "The user intended FP parameter/result passing to conform to AAPCS, VFP variant",
    2: "The user intended FP parameter/result passing to conform to tool chain-specific conventions",
    3: "Code is compatible with both the base and VFP variants; the user did not permit non-variadic functions to pass FP parameters/results",
}

Tag_ABI_WMMX_args = {
    0: "The user intended WMMX parameter/result passing conform to the AAPCS, base variant",
    1: "The user intended WMMX parameter/result passing conform to Intels WMMX conventions",
    2: "The user intended WMMX parameter/result passing conforms to tool chain-specific conventions",
}

Tag_ABI_optimization_goals = {
    0: "No particular optimization goals, or no information recorded",
    1: "Optimized for speed, but small size and good debug illusion preserved",
    2: "Optimized aggressively for speed, small size and debug illusion sacrificed",
    3: "Optimized for small size, but speed and debugging illusion preserved",
    4: "Optimized aggressively for small size, speed and debug illusion sacrificed",
    5: "Optimized for good debugging, but speed and small size preserved",
    6: "Optimized for best debugging illusion, speed and small size sacrificed",
}

Tag_ABI_FP_optimization_goals = {
    0: "No particular FP optimization goals, or no information recorded",
    1: "Optimized for speed, but small size and good accuracy preserved",
    2: "Optimized aggressively for speed, small size and accuracy sacrificed",
    3: "Optimized for small size, but speed and accuracy preserved",
    4: "Optimized aggressively for small size, speed and accuracy sacrificed",
    5: "Optimized for accuracy, but speed and small size preserved",
    6: "Optimized for best accuracy, speed and small size sacrificed",
}



ATTRIBUTES = {
    1:  AttributeDescription("Tag_File", 1, Reader.INT32, Ident),
    2:  AttributeDescription("Tag_Section", 2, Reader.INT32, Ident),
    3:  AttributeDescription("Tag_Symbol", 3, Reader.INT32, Ident),
    4:  AttributeDescription("Tag_CPU_raw_name", 4, Reader.CSTRING, Ident),
    5:  AttributeDescription("Tag_CPU_name", 5, Reader.CSTRING, Ident),
    6:  AttributeDescription("Tag_CPU_arch", 6, Reader.ULEB, Tag_CPU_arch),
    7:  AttributeDescription("Tag_CPU_arch_profile", 7, Reader.ULEB, Tag_CPU_arch_profile),
    8:  AttributeDescription("Tag_ARM_ISA_use", 8, Reader.ULEB, Tag_ARM_ISA_use),
    9:  AttributeDescription("Tag_THUMB_ISA_use", 9, Reader.ULEB, Tag_THUMB_ISA_use),
    10: AttributeDescription("Tag_FP_arch", 10, Reader.ULEB, Tag_FP_arch),
    11: AttributeDescription("Tag_WMMX_arch", 11, Reader.ULEB, Tag_WMMX_arch),
    12: AttributeDescription("Tag_Advanced_SIMD_arch", 12, Reader.ULEB, Tag_Advanced_SIMD_arch),
    13: AttributeDescription("Tag_PCS_config", 13, Reader.ULEB, Tag_PCS_config),
    14: AttributeDescription("Tag_ABI_PCS_R9_use",  14, Reader.ULEB, Tag_ABI_PCS_R9_use),
    15: AttributeDescription("Tag_ABI_PCS_RW_data", 15, Reader.ULEB,Tag_ABI_PCS_RW_data),
    16: AttributeDescription("Tag_ABI_PCS_RO_data", 16, Reader.ULEB, Tag_ABI_PCS_RO_data),
    17: AttributeDescription("Tag_ABI_PCS_GOT_use", 17, Reader.ULEB, Tag_ABI_PCS_GOT_use),
    18: AttributeDescription("Tag_ABI_PCS_wchar_t", 18, Reader.ULEB, Tag_ABI_PCS_wchar_t),
    19: AttributeDescription("Tag_ABI_FP_rounding", 19, Reader.ULEB, Tag_ABI_FP_rounding),
    20: AttributeDescription("Tag_ABI_FP_denormal", 20, Reader.ULEB, Tag_ABI_FP_rounding),
    21: AttributeDescription("Tag_ABI_FP_exceptions", 21, Reader.ULEB, Tag_ABI_FP_exceptions),
    22: AttributeDescription("Tag_ABI_FP_user_exceptions", 22, Reader.ULEB, Tag_ABI_FP_user_exceptions),
    23: AttributeDescription("Tag_ABI_FP_number_model", 23, Reader.ULEB, Tag_ABI_FP_number_model),
    24: AttributeDescription("Tag_ABI_align_needed", 24, Reader.ULEB, Tag_ABI_align_needed),
    25: AttributeDescription("Tag_ABI_align8_preserved", 25, Reader.ULEB, Tag_ABI_align_preserved),
    26: AttributeDescription("Tag_ABI_enum_size", 26, Reader.ULEB, Tag_ABI_enum_size),
    27: AttributeDescription("Tag_ABI_HardFP_use", 27, Reader.ULEB, Tag_ABI_HardFP_use),
    28: AttributeDescription("Tag_ABI_VFP_args", 28, Reader.ULEB, Tag_ABI_VFP_args),
    29: AttributeDescription("Tag_ABI_WMMX_args", 29, Reader.ULEB, Tag_ABI_WMMX_args),
    30: AttributeDescription("Tag_ABI_optimization_goals", 30, Reader.ULEB, Tag_ABI_optimization_goals),
    31: AttributeDescription("Tag_ABI_FP_optimization_goals", 31, Reader.ULEB, Tag_ABI_FP_optimization_goals),
    32: AttributeDescription("Tag_compatibility", 32, Reader.CSTRING, Ident),
    34: AttributeDescription("Tag_CPU_unaligned_access", 34, Reader.ULEB, Tag_CPU_unaligned_access),
    36: AttributeDescription("Tag_FP_HP_extension (was Tag_VFP_HP_extension)",  36, Reader.ULEB, Tag_FP_HP_extension),
    38: AttributeDescription("Tag_ABI_FP_16bit_format", 38, Reader.ULEB, Tag_ABI_FP_16bit_format),
    42: AttributeDescription("Tag_MPextension_use", 42, Reader.ULEB, Tag_MPextension_use),
    44: AttributeDescription("Tag_DIV_use", 44, Reader.ULEB, Tag_DIV_use),
    64: AttributeDescription("Tag_nodefaults", 64, Reader.ULEB, Ident),
    65: AttributeDescription("Tag_also_compatible_with", 65, Reader.CSTRING, Ident),
    67: AttributeDescription("Tag_conformance", 67, Reader.CSTRING, Ident),
    66: AttributeDescription("Tag_T2EE_use", 66, Reader.ULEB, Tag_T2EE_use),
    68: AttributeDescription("Tag_Virtualization_use", 68, Reader.ULEB, Tag_Virtualization_use),
    70: AttributeDescription("Tag_MPextension_use", 70, Reader.ULEB, Tag_MPextension_use),
}

"""
ADI         Analog Devices
acle        Reserved for use by Arm C Language Extensions.
aeabi       Reserved to the ABI for the Arm Architecture (EABI pseudo-vendor)
AnonXyz
anonXyz     Reserved to private experiments by the Xyz vendor. Guaranteed not to clash with any registered vendor name.
ARM Arm     Ltd (Note: the company, not the processor).
cxa         C++ ABI pseudo-vendor
FSL         Freescale Semiconductor Inc.
GHS         Green Hills Systems
gnu         GNU compilers and tools (Free Software Foundation)
iar         IAR Systems
icc         ImageCraft Creations Inc (ImageCraft C Compiler)
intel       Intel Corporation
ixs         Intel Xscale
llvm        The LLVM/Clang projects
PSI         PalmSource Inc.
RAL         Rowley Associates Ltd
SEGGER      SEGGER Microcontroller GmbH
somn        SOMNIUM Technologies Limited.
TASKING     Altium Ltd.
TI          TI Inc.
tls         Reserved for use in thread-local storage routines.
WRS         Wind River Systems.
"""

"""
<format-version>
    [ <section-length> "vendor-name"
        [ <file-tag> <size> <attribute>*
        | <section-tag> <size> <section-number>* 0 <attribute>*
        | <symbol-tag> <size> <symbol-number>* 0 <attribute>*
        ]+
]*
"""

class Attribute:

    def __init__(self, tag, tag_name, value, description = None):
        self.tag = tag
        self.tag_name = tag_name
        self.value = value
        if description is None:
            self.description = value

    def __repr__(self):
        return "Attribute(tag = {}, tag_name = {}, value = {}, description = {})".format(
            self.tag, self.tag_name, self.value, self.description
        )


def parse(buffer, byteorder = "<"):
    """
    Parameters
    ----------
    buffer: bytes-like

    byteorder: char
        "<": Little-endian
        ">": Big-endian

    Returns
    -------
    dict
        key: Vendor name
        values: list of attributes
    """
    Integer = Int32ul if byteorder == "<" else Int32ub
    Section = Struct(
        "len" / Integer,
        "vendor" / CString(encoding = "ascii"),
        "_pos" / Tell,
        "data" / Bytes(this.len - this._pos),
    )
    SubSectionHeader = Struct(
        "tag" / Byte,
        "len" / Integer,
        "_pos" / Tell,
        "data" / Bytes(this.len - this._pos),
    )

    RawAttribute = Struct(
        "tag" / ULEB,
        "parameterType" / Computed(lambda ctx: ATTRIBUTES[ctx.tag].parameterType),
        "name" / Computed(lambda ctx: ATTRIBUTES[ctx.tag].tag),
        "_conv" / Computed(lambda ctx: ATTRIBUTES[ctx.tag].conv),
        "value" / Switch(this.parameterType, { Reader.INT32: Integer, Reader.ULEB: ULEB, Reader.CSTRING: CString(encoding = "ascii") }),
        "pos" / Tell,
    )
    format_version = buffer[0]
    i = 1
    length = len(buffer)
    result = {}
    while True:
        section = Section.parse(buffer[i : ])
        if not section.vendor in result:
            result[section.vendor] = []
        i += section.len
        res = SubSectionHeader.parse(section.data)
        j = 0
        while j < len(res.data):
            attr = RawAttribute.parse(res.data[j : ])
            r = Attribute(attr.tag, attr.name, attr.value)
            if attr._conv != Ident:
                r.description = attr._conv[attr.value]
            result[section.vendor].append(r)
            j += attr.pos
        if i >= length:
            break
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import objutils.hexfile as hexfile

DATA                        = 0
EOF                         = 1
EXTENDED_SEGMENT_ADDRESS    = 2
START_SEGMENT_ADDRESS       = 3
EXTENDED_LINEAR_ADDRESS     = 4
START_LINEAR_ADDRESS        = 5


class Codec(object):
    """

    """

    def __init__(self, file_like):
        self.file_like = file_like

    def readlines(self):
        for line in self.file_like.readlines():
            yield line

    def writelines(self, lines):
        for line in lines:
            self.file_like.write(line)


class Reader(hexfile.Reader):
    """

    """

    FORMAT_SPEC = (
        (hexfile.TYPE_FROM_RECORD, "LL AAAA:DD CCCC"),
    )
    DATA_SEPARATOR =  " "

    def check_line(self, line, format_type):
        if line.length != len(line.chunk):
            raise hexfile.InvalidRecordLengthError("Byte count doesn't match length of actual data.")
        # todo: factor out checksum calculation from line!!!
        checksum = (sum(line.chunk) & 0xffff)
        if line.checksum != checksum:
            raise hexfile.InvalidRecordChecksumError()

    def is_data_line(self, line, format_type):
        return True


class Writer(hexfile.Writer):
    """

    """

    MAX_ADDRESS_BITS = 16

    def compose_row(self, address, length, row):
        checksum = sum(row) % 65536
        return "{0:02X} {1:04X}:{2!s} {3:04X}".format(length, address, Writer.hex_bytes(row, spaced = True), checksum)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import objutils.hexfile as hexfile
import objutils.utils as utils
import objutils.checksums as checksums

DATA    = 1
EOF     = 2


class Reader(hexfile.Reader):

    FORMAT_SPEC = (
        (DATA,  ";LLAAAADDCCCC"),
        (EOF,   ";00")
    )

    def check_line(self, line, format_type):
        if format_type == DATA:
            if line.length != len(line.chunk):
                raise hexfile.InvalidRecordLengthError("Byte count doesn't match length of actual data.")
            checksum = checksums.lrc(utils.make_list(utils.int_to_array(line.address), line.length, line.chunk), 16, checksums.COMPLEMENT_NONE)
            if line.checksum != checksum:
                raise hexfile.InvalidRecordChecksumError()

    def is_data_line(self, line, format_type):
        return format_type == DATA


class Writer(hexfile.Writer):

    MAX_ADDRESS_BITS = 16

    def compose_row(self, address, length, row):
        checksum = checksums.lrc(utils.make_list(utils.int_to_array(address), length, row), 16, checksums.COMPLEMENT_NONE)
        line = ";{0:02X}{1:04X}{2!s}{3:04X}".format(length, address, Writer.hex_bytes(row), checksum)
        return line

    def compose_footer(self, meta):
        return ";00"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Reader/Writer S Hexdump Format (rfc4149).
"""

__version__ = "0.1.1"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from hashlib import sha1
import re
import xml.etree.ElementTree as ET

from objutils.logger import Logger
from objutils.utils import create_string_buffer, PYTHON_VERSION
from objutils.image import Image
from objutils.section import Section

SHF_DTD = """<!--
    DTD for the S Hexdump Format, as of 2003-10-10
    Linus Walleij, Joachim Strombergson, Patrik Faltstrom 2003

    Refer to this DTD as:

    <!ENTITY % SHF PUBLIC "-//IETF//DTD SHF//EN" "http://ietf.org/dtd/shf.dtd">
        %SHF;
-->

<?xml version="1.0" encoding="UTF-8"?>

<!ELEMENT dump (block)+>
<!ATTLIST dump
    name CDATA #REQUIRED
    blocks CDATA #IMPLIED>

<!ELEMENT block (#PCDATA)>
    <!ATTLIST block
    name CDATA #REQUIRED
    address CDATA #REQUIRED
    word_size CDATA #REQUIRED
    length CDATA #REQUIRED
    checksum CDATA #REQUIRED>
"""

SHA1_DIGEST = lambda text: sha1(text).hexdigest()

WHITESPACE = re.compile("\s*")

remove_ws = lambda text: WHITESPACE.sub("", text)

class Reader(object):
    """
    """

    logger = Logger(__name__)

    def load(self, fp):
        if isinstance(fp, str):
            fp = open(fp, "rb")
        data = fp.read()
        root = ET.fromstring(data)
        sections = []
        for idx, child in enumerate(root):
            tag = child.tag
            attrib = child.attrib
            text = remove_ws(child.text)
            section_data = bytearray.fromhex(text)
            name = attrib.get('name')
            if name is None:
                self.logger.error("Block #{}: Missing required attribute `name`.".format(idx))
                continue
            address = attrib.get('address')
            if address:
                address = remove_ws(address)
                address = int(address, 16)
            else:
                self.logger.error("Block #{}: Missing required attribute `address`.".format(idx))
                continue
            length = attrib.get('length')
            if length:
                length = remove_ws(length)
                length = int(length, 16)
            else:
                self.logger.error("Block #{}: Missing required attribute `length`.".format(idx))
                continue
            word_size = attrib.get('word_size')
            if word_size:
                word_size = remove_ws(word_size)
                word_size = int(word_size, 16)
            else:
                self.logger.error("Block #{}: Missing required attribute `wordsize`.".format(idx))
                continue
            if len(section_data) != (length * word_size):
                self.logger.error("Block #{}: Mismatch between (`length` * `word_size`) and actual block length.".format(idx))
                continue
            checksum = attrib.get('checksum')
            if checksum:
                checksum = remove_ws(checksum)
                if SHA1_DIGEST(section_data) != checksum:
                    self.logger.error("Block #{}: Wrong `checksum`.".format(idx))
                    continue
            else:
                self.logger.error("Block #{}: Missing required attribute `checksum`.".format(idx))
                continue
            #print(tag, attrib)
            #print(section_data, SHA1_DIGEST(section_data))
            sections.append(Section(address, section_data))
        img = Image(sections)
        if hasattr(fp, "close"):
            fp.close()
        return img

    def loads(self, image):
        if PYTHON_VERSION.major == 3:
            if isinstance(image, str):
                return self.load(create_string_buffer(bytes(image, "ascii")))
            else:
                return self.load(create_string_buffer(image))
        else:
            return self.load(create_string_buffer(image))


class Writer(object):
    """
    """

    logger = Logger(__name__)

    def dump(self, fp, image, **kws):
        if isinstance(fp, str):
            fp = open(fp, "wb")
        fp.write(self.dumps(image))
        if hasattr(fp, "close"):
            fp.close()

    def dumps(self, image, **kws):
        BLOCK_SIZE = 16
        result = []
        result.append('<?xml version="1.0" encoding="UTF-8"?>')
        result.append('<dump name="SHF dump by objutils" blocks="{:04x}">'.format(len(image._sections)))
        if hasattr(image, "sections") and  not image.sections:
            return b''
        sections = sorted(image.sections, key = lambda x: x.start_address)
        for idx, section in enumerate(sections):
            result.append('    <block name="Section #{:04x}" address="{:08x}" word_size="01" length="{:08x}" checksum="{}">'.\
                format(idx, section.start_address, section.length, SHA1_DIGEST(section.data)))
            nblocks = len(section.data) // BLOCK_SIZE
            remaining = len(section.data) % BLOCK_SIZE
            offset = 0
            for _ in range(nblocks):
                result.append("        {}".format(" ".join(
                    ["{:02x}".format(x) for x in section.data[offset : offset + BLOCK_SIZE]]))
                )
                offset += BLOCK_SIZE
            if remaining:
                result.append("        {}".format(" ".join(
                    ["{:02x}".format(x) for x in section.data[offset : offset + remaining]]))
                )
            result.append('    </block>')
        result.append("</dump>")
        return bytes("\n".join(result), encoding = "ascii")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import os
import struct


class PlainBinaryReader(object):
    LITTLE_ENDIAN   = '<'
    BIG_ENDIAN      = '>'

    def __init__(self, image, byte_order_prefix = "@"):
        self.image = image
        self.image.seek(0, os.SEEK_END)
        self._size = self.image.tell()
        self.image.seek(0, os.SEEK_SET)
        self.byte_order_prefix = byte_order_prefix
        self.pos = 0

    def _get_pos(self):
        return self.image.tell()

    def _set_pos(self, pos):
        self.image.seek(pos, os.SEEK_SET)

    def _get_size(self):
        return self._size

    def reset(self):
        self.pos = 0

    def next_byte(self):
        return self.u8()

    def value(self, conversion_code, size):
        res, = struct.unpack('%c%c' % (self.byte_order_prefix, conversion_code, ), self.image.read(size))
        return res

    def u8(self):
        return self.value('B', 1)

    def u16(self):
        return self.value('H', 2)

    def u32(self):
        return self.value('L', 4)

    def u64(self):
        return self.value('Q', 8)

    def s8(self):
        return self.value('b', 1)

    def s16(self):
        return self.value('h', 2)

    def s32(self):
        return self.value('l', 4)

    def s64(self):
        return self.value('q', 8)

    def uleb(self):
        result = 0
        shift = 0
        while True:
            bval = self.next_byte()
            result |= ((bval & 0x7f) << shift)
            if bval & 0x80 == 0:
                break
            shift += 7
        return result

    def sleb(self):
        result = 0
        shift = 0
        idx =0
        while True:
            bval = self.next_byte()
            result |= ((bval & 0x7f) << shift)
            shift += 7
            idx += 1
            if bval & 0x80 == 0:
                break
        if (shift < 32) or (bval & 0x40) == 0x40:
            mask = - (1 << (idx * 7))
            result |= mask
        return result

    def asciiz(self):
        result = []
        while True:
            ch = self.next_byte()
            if ch == 0:
                break
            else:
                result.append(ch)
        return ''.join(chr(c) for c in result)

    pos = property(_get_pos, _set_pos)
    size = property(_get_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import enum
import operator
import functools
import struct
import zlib

COMPLEMENT_NONE = 0
COMPLEMENT_ONES = 1
COMPLEMENT_TWOS = 2

def lrc(data, width, comp = COMPLEMENT_NONE):
    """Longitudinal redundancy check.
    """

    mask = (2 ** width)

    cs = sum(data) % mask

    if comp == COMPLEMENT_NONE:
        pass
    elif comp == COMPLEMENT_ONES:
        cs ^= (mask - 1)
    elif comp == COMPLEMENT_TWOS:
        cs = ((cs ^ (mask - 1)) + 1) % mask

    return cs


def rolb(value):
    """Rotate byte left.
    """
    value &= 0xff
    carry = (value & 0x80) == 0x80
    value = (value << 1) & 0xff
    value |= 1 if carry else 0
    return value

def rorb(value):
    """Rotate byte right.
    """
    value &= 0xff
    carry = (value & 0x01) == 0x01
    value = value >> 1
    value |= 0x80 if carry else 0
    return value

def xor(frame, invert = True):
    return functools.reduce(operator.xor, frame, 0xff if invert else 0x00)

ROTATE_LEFT = rolb
ROTATE_RIGHT = rorb

def rotatedXOR(values, width, rotator):
    """Rotated XOR cipher.
    """
    cs = 0
    for value in values:
        cs ^= value
        cs = rotator(cs)
    return cs  % (2 ** width)

def nibble_sum(data):
    result = 0
    for d in data:
        hn = (d & 0xf0) >> 4
        ln = d & 0x0f
        result +=  hn + ln
    return result % 256

##
##
##
class Algorithm(enum.IntEnum):
    """Enumerates available checksum algorithms
    """
    CHK_ADD_11 = 1
    CHK_ADD_12 = 2
    CHK_ADD_14 = 3
    CHK_ADD_22 = 4
    CHK_ADD_24 = 5
    CHK_ADD_44 = 6
    CHK_CRC_16 = 7
    CHK_CRC_16_CCITT = 8
    CHK_CRC_32 = 9
    CHK_USER_DEFINED = 10


CRC16 = (
    0x0000, 0x8005, 0x800F, 0x000A, 0x801B, 0x001E, 0x0014, 0x8011,
    0x8033, 0x0036, 0x003C, 0x8039, 0x0028, 0x802D, 0x8027, 0x0022,
    0x8063, 0x0066, 0x006C, 0x8069, 0x0078, 0x807D, 0x8077, 0x0072,
    0x0050, 0x8055, 0x805F, 0x005A, 0x804B, 0x004E, 0x0044, 0x8041,
    0x80C3, 0x00C6, 0x00CC, 0x80C9, 0x00D8, 0x80DD, 0x80D7, 0x00D2,
    0x00F0, 0x80F5, 0x80FF, 0x00FA, 0x80EB, 0x00EE, 0x00E4, 0x80E1,
    0x00A0, 0x80A5, 0x80AF, 0x00AA, 0x80BB, 0x00BE, 0x00B4, 0x80B1,
    0x8093, 0x0096, 0x009C, 0x8099, 0x0088, 0x808D, 0x8087, 0x0082,
    0x8183, 0x0186, 0x018C, 0x8189, 0x0198, 0x819D, 0x8197, 0x0192,
    0x01B0, 0x81B5, 0x81BF, 0x01BA, 0x81AB, 0x01AE, 0x01A4, 0x81A1,
    0x01E0, 0x81E5, 0x81EF, 0x01EA, 0x81FB, 0x01FE, 0x01F4, 0x81F1,
    0x81D3, 0x01D6, 0x01DC, 0x81D9, 0x01C8, 0x81CD, 0x81C7, 0x01C2,
    0x0140, 0x8145, 0x814F, 0x014A, 0x815B, 0x015E, 0x0154, 0x8151,
    0x8173, 0x0176, 0x017C, 0x8179, 0x0168, 0x816D, 0x8167, 0x0162,
    0x8123, 0x0126, 0x012C, 0x8129, 0x0138, 0x813D, 0x8137, 0x0132,
    0x0110, 0x8115, 0x811F, 0x011A, 0x810B, 0x010E, 0x0104, 0x8101,
    0x8303, 0x0306, 0x030C, 0x8309, 0x0318, 0x831D, 0x8317, 0x0312,
    0x0330, 0x8335, 0x833F, 0x033A, 0x832B, 0x032E, 0x0324, 0x8321,
    0x0360, 0x8365, 0x836F, 0x036A, 0x837B, 0x037E, 0x0374, 0x8371,
    0x8353, 0x0356, 0x035C, 0x8359, 0x0348, 0x834D, 0x8347, 0x0342,
    0x03C0, 0x83C5, 0x83CF, 0x03CA, 0x83DB, 0x03DE, 0x03D4, 0x83D1,
    0x83F3, 0x03F6, 0x03FC, 0x83F9, 0x03E8, 0x83ED, 0x83E7, 0x03E2,
    0x83A3, 0x03A6, 0x03AC, 0x83A9, 0x03B8, 0x83BD, 0x83B7, 0x03B2,
    0x0390, 0x8395, 0x839F, 0x039A, 0x838B, 0x038E, 0x0384, 0x8381,
    0x0280, 0x8285, 0x828F, 0x028A, 0x829B, 0x029E, 0x0294, 0x8291,
    0x82B3, 0x02B6, 0x02BC, 0x82B9, 0x02A8, 0x82AD, 0x82A7, 0x02A2,
    0x82E3, 0x02E6, 0x02EC, 0x82E9, 0x02F8, 0x82FD, 0x82F7, 0x02F2,
    0x02D0, 0x82D5, 0x82DF, 0x02DA, 0x82CB, 0x02CE, 0x02C4, 0x82C1,
    0x8243, 0x0246, 0x024C, 0x8249, 0x0258, 0x825D, 0x8257, 0x0252,
    0x0270, 0x8275, 0x827F, 0x027A, 0x826B, 0x026E, 0x0264, 0x8261,
    0x0220, 0x8225, 0x822F, 0x022A, 0x823B, 0x023E, 0x0234, 0x8231,
    0x8213, 0x0216, 0x021C, 0x8219, 0x0208, 0x820D, 0x8207, 0x0202,
)

CRC16_CCITT = (
    0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50A5, 0x60C6, 0x70E7,
    0x8108, 0x9129, 0xA14A, 0xB16B, 0xC18C, 0xD1AD, 0xE1CE, 0xF1EF,
    0x1231, 0x0210, 0x3273, 0x2252, 0x52B5, 0x4294, 0x72F7, 0x62D6,
    0x9339, 0x8318, 0xB37B, 0xA35A, 0xD3BD, 0xC39C, 0xF3FF, 0xE3DE,
    0x2462, 0x3443, 0x0420, 0x1401, 0x64E6, 0x74C7, 0x44A4, 0x5485,
    0xA56A, 0xB54B, 0x8528, 0x9509, 0xE5EE, 0xF5CF, 0xC5AC, 0xD58D,
    0x3653, 0x2672, 0x1611, 0x0630, 0x76D7, 0x66F6, 0x5695, 0x46B4,
    0xB75B, 0xA77A, 0x9719, 0x8738, 0xF7DF, 0xE7FE, 0xD79D, 0xC7BC,
    0x48C4, 0x58E5, 0x6886, 0x78A7, 0x0840, 0x1861, 0x2802, 0x3823,
    0xC9CC, 0xD9ED, 0xE98E, 0xF9AF, 0x8948, 0x9969, 0xA90A, 0xB92B,
    0x5AF5, 0x4AD4, 0x7AB7, 0x6A96, 0x1A71, 0x0A50, 0x3A33, 0x2A12,
    0xDBFD, 0xCBDC, 0xFBBF, 0xEB9E, 0x9B79, 0x8B58, 0xBB3B, 0xAB1A,
    0x6CA6, 0x7C87, 0x4CE4, 0x5CC5, 0x2C22, 0x3C03, 0x0C60, 0x1C41,
    0xEDAE, 0xFD8F, 0xCDEC, 0xDDCD, 0xAD2A, 0xBD0B, 0x8D68, 0x9D49,
    0x7E97, 0x6EB6, 0x5ED5, 0x4EF4, 0x3E13, 0x2E32, 0x1E51, 0x0E70,
    0xFF9F, 0xEFBE, 0xDFDD, 0xCFFC, 0xBF1B, 0xAF3A, 0x9F59, 0x8F78,
    0x9188, 0x81A9, 0xB1CA, 0xA1EB, 0xD10C, 0xC12D, 0xF14E, 0xE16F,
    0x1080, 0x00A1, 0x30C2, 0x20E3, 0x5004, 0x4025, 0x7046, 0x6067,
    0x83B9, 0x9398, 0xA3FB, 0xB3DA, 0xC33D, 0xD31C, 0xE37F, 0xF35E,
    0x02B1, 0x1290, 0x22F3, 0x32D2, 0x4235, 0x5214, 0x6277, 0x7256,
    0xB5EA, 0xA5CB, 0x95A8, 0x8589, 0xF56E, 0xE54F, 0xD52C, 0xC50D,
    0x34E2, 0x24C3, 0x14A0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
    0xA7DB, 0xB7FA, 0x8799, 0x97B8, 0xE75F, 0xF77E, 0xC71D, 0xD73C,
    0x26D3, 0x36F2, 0x0691, 0x16B0, 0x6657, 0x7676, 0x4615, 0x5634,
    0xD94C, 0xC96D, 0xF90E, 0xE92F, 0x99C8, 0x89E9, 0xB98A, 0xA9AB,
    0x5844, 0x4865, 0x7806, 0x6827, 0x18C0, 0x08E1, 0x3882, 0x28A3,
    0xCB7D, 0xDB5C, 0xEB3F, 0xFB1E, 0x8BF9, 0x9BD8, 0xABBB, 0xBB9A,
    0x4A75, 0x5A54, 0x6A37, 0x7A16, 0x0AF1, 0x1AD0, 0x2AB3, 0x3A92,
    0xFD2E, 0xED0F, 0xDD6C, 0xCD4D, 0xBDAA, 0xAD8B, 0x9DE8, 0x8DC9,
    0x7C26, 0x6C07, 0x5C64, 0x4C45, 0x3CA2, 0x2C83, 0x1CE0, 0x0CC1,
    0xEF1F, 0xFF3E, 0xCF5D, 0xDF7C, 0xAF9B, 0xBFBA, 0x8FD9, 0x9FF8,
    0x6E17, 0x7E36, 0x4E55, 0x5E74, 0x2E93, 0x3EB2, 0x0ED1, 0x1EF0,
)


def reflect(data, nBits):
    """Reflect data, i.e. reverse bit order.

    Parameters
    ----------
    data : int
    nBits : int
        width in bits of `data`
    """
    reflection = 0x00000000
    for bit in range(nBits):
        if data & 0x01:
            reflection |= (1 << ((nBits - 1) - bit))
        data = (data >> 1)
    return reflection


class Crc16:
    """Calculate CRC (16-bit)


    Parameters
    ----------
    table: list-like
        lookup table for CRC calculation
    initalRemainder : int
        value to start with
    finalXorValue : int
        final XOR value
    reflectData : bool
        reflect input data
    reflectRemainder : bool
        reflect output data

    .. [1] A PAINLESS GUIDE TO CRC ERROR DETECTION ALGORITHMS
           http://www.ross.net/crc/download/crc_v3.txt
    .. [2] Understanding and implementing CRC (Cyclic Redundancy Check)
           calculation
           http://www.sunshine2k.de/articles/coding/crc/understanding_crc.html
    .. [3] Online CRC calculator
           http://zorc.breitbandkatze.de/crc.html
    """
    WIDTH = 16

    def __init__(self, table, initalRemainder, finalXorValue, reflectData,
                 reflectRemainder):
        self.table = table
        self.initalRemainder = initalRemainder
        self.finalXorValue = finalXorValue
        self.reflectData = reflectData
        self.reflectRemainder = reflectRemainder

    def __call__(self, frame):
        remainder = self.initalRemainder
        for ch in frame:
            data = self.reflectIn(ch, remainder)
            remainder = (self.table[data] ^ (remainder << 8)) & 0xffff
        return self.reflectOut(remainder)

    def reflectIn(self, ch, remainder):
        if self.reflectData:
            return (reflect(ch, 8) ^ (remainder >> (self.WIDTH - 8))) & 0xff
        else:
            return (ch ^ (remainder >> (self.WIDTH - 8))) & 0xff

    def reflectOut(self, remainder):
        if self.reflectRemainder:
            return reflect(remainder, 16) ^ self.finalXorValue
        else:
            return remainder ^ self.finalXorValue


def adder(modulus):
    """Factory function for modulus adders

    Parameters
    ----------
    modulus : int
        modulus to use

    Returns
    -------
    function
        adder function

    Examples
    --------
    >>> a256=adder(256)
    >>> a256([11, 22, 33, 44, 55, 66, 77, 88, 99])
    239

    """
    def add(frame):
        return sum(frame) % modulus
    return add


def wordSum(modulus, step):
    """Factory function for (double-)word modulus sums

    Parameters
    ----------
    modulus : int
    step : [2, 4]
        2 - word wise
        4 - double-word wise

    Returns
    -------
    function
        summation function
    """
    def add(frame):
        if step == 2:
            mask = "<H"
        elif step == 4:
            mask = "<I"
        else:
            raise NotImplementedError("Only WORDs or DWORDs are supported.")
        x = [struct.unpack(mask, frame[x:x + step])[0]
             for x in range(0, len(frame), step)]
        return sum(x) % modulus
    return add


ADD11 = adder(2 ** 8)
ADD12 = adder(2 ** 16)
ADD14 = adder(2 ** 32)
ADD22 = wordSum(2 ** 16, 2)
ADD24 = wordSum(2 ** 32, 2)
ADD44 = wordSum(2 ** 32, 4)
CRC16 = Crc16(CRC16, 0x0000, 0x0000, True, True)
CRC16_CCITT = Crc16(CRC16_CCITT, 0xffff, 0x0000, False, False)


def CRC32(x):
    return zlib.crc32(bytes(x)) & 0xffffffff


def userDefined(x):
    """User defined algorithms are not supported yet.
    """
    raise NotImplementedError(
        "Checksum method 'CHK_USER_DEFINED' not supported yet.")


ALGO = {
    "CHK_ADD_11":       ADD11,
    "CHK_ADD_12":       ADD12,
    "CHK_ADD_14":       ADD14,
    "CHK_ADD_22":       ADD22,
    "CHK_ADD_24":       ADD24,
    "CHK_ADD_44":       ADD44,
    "CHK_CRC_16":       CRC16,
    "CHK_CRC_16_CCITT":  CRC16_CCITT,
    "CHK_CRC_32":       CRC32,
    "CHK_USER_DEFINED": userDefined
}


def check(frame, algo):
    """Calculate checksum using given algorithm

    Parameters
    ----------
    frame : list of integers
    algo : `ALGO`

    Returns
    -------
    int
    """
    fun = ALGO.get(algo)
    if fun:
        return fun(frame)
    else:
        raise NotImplementedError("Invalid algorithm '{}'.".format(algo))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""

"""

from __future__ import print_function

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from collections import namedtuple
import enum
from operator import attrgetter, eq
import sys

from sortedcontainers import SortedList

from objutils.section import Section, join_sections
from objutils.exceptions import InvalidAddressError

SearchType = namedtuple("SearchType", "start_address")


#
# TODO: Use crypto hashes (comparison, optimized storage, ...)
#

class AddressSpace(enum.IntEnum):
    """Adress-space constants.
    """
    AS_16   = 0
    AS_24   = 1
    AS_32   = 2
    AS_64   = 3


class Image(object):
    """Manage images.

    An image is a collection of :class:`Section`s and meta-data.

    Parameters
    ----------
    sections: iteratable (typically `list`or `tuple`
        The sections the image should initialized with.
    meta: object
        Arbitrary meta-data.
    """

    def __init__(self, sections = None, join = True, meta = None):
        if meta is None:
            meta = {}
        if not sections:
            sections = []
        elif isinstance(sections, Section) or hasattr(sections, "__iter__"):
            sections = list(sections)
        else:
            raise TypeError("Argument section is of wrong type '{}'".format(sections))
        self._sections = SortedList(sections, key = attrgetter("start_address"))
        self._join = join
        if join:
            self.join_sections()
        _validate_sections(self._sections)
        self.address = 0
        #if meta and not isinstance(meta, MetaRecord):
        #    raise TypeError("meta-data must be of instance 'MetaRecord'")
        self.meta = meta

    def __repr__(self):
        result = []
        for segment in self.sections:
            result.append(repr(segment))
        return '\n'.join(result)

    __str__ = __repr__

    def __len__(self):
        return sum(len(s) for s in self.sections)

    def __iter__(self):
        return iter(self.sections)

    def __getitem__(self, idx):
        return self.sections[idx]

    def __eq__(self, other):
        if len(self.sections) == len(other.sections):
            return all(eq(l, r) for l, r in zip(self.sections, other.sections))
        else:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __contains__(self, addr):
        return any(addr in sec for sec in self.sections)

    def hexdump(self, fp = sys.stdout):
        """

        """
        for idx, section in enumerate(self.sections):
            print("\nSection #{0:04d}".format(idx ), file = fp)
            print("-" * 13, file = fp)
            section.hexdump(fp)

    def _call_address_function(self, func_name, addr, *args, **kws):
        for section in self.sections:
            if addr in section:
                func = getattr(section, func_name)
                return func(addr, *args, **kws)
        raise InvalidAddressError("Address 0x{:08x} not in range.".format(addr))

    def read(self, addr, length, **kws):
        """Read bytes from image.

        Parameters
        ----------
        addr: int
            Startaddress.

        length: int
            Number of bytes to read.

        Returns
        -------
        bytes

        Raises
        ------
        :class:`InvalidAddressError`
            if `addr` is out of range

        Note
        ----
            if `addr` + `len` is out of range, result is silently truncated, i.e. without raising an exception.
        """
        return self._call_address_function("read", addr, length, **kws)

    def write(self, addr, data, **kws):
        """Write bytes to image.

        Parameters
        ----------
        addr: int
            Startaddress.

        data: bytes

        Raises
        ------
        :class:`InvalidAddressError`
            if `addr` is out of range
        """
        self._call_address_function("write", addr, data, **kws)

    def read_numeric(self, addr, dtype, **kws):
        """

        """
        return self._call_address_function("read_numeric", addr, dtype, **kws)

    def write_numeric(self, addr, value, dtype, **kws):
        """
        """
        self._call_address_function("write_numeric", addr, value, dtype, **kws)

    def read_numeric_array(self, addr, length, dtype, **kws):
        """

        """
        return self._call_address_function("read_numeric_array", addr, length, dtype, **kws)

    def write_numeric_array(self, addr, data, dtype, **kws):# TODO: bounds-checking.
        """
        """
        self._call_address_function("write_numeric_array", addr, data, dtype)

    def write_ndarray(self, addr, array, order = None, **kws):
        """
        """
        self._call_address_function("write_ndarray", addr, array, order = None, **kws)

    def read_ndarray(self, addr, length, dtype, shape = None, order = None, **kws):
        """
        """
        return self._call_address_function("read_ndarray", addr, length, dtype, shape, order, **kws)


    def read_string(self, addr, encoding = "latin1", length = -1, **kws):
        """

        """
        return self._call_address_function("read_string", addr, encoding, length, **kws)

    def write_string(self, addr, value, encoding = "latin1", **kws):
        """

        """
        self._call_address_function("write_string", addr, value, encoding, **kws)


    def _address_contained(self, address, length):
        """Check if address space exists.

        Parameters
        ----------
        address: int
        length: int

        Returns
        -------
        bool
        """
        return address in self or (address + length - 1) in self

    def insert_section(self, data, start_address = None, join = True):
        """Insert/add a new section to image.

        Parameters
        ----------
        data: convertible to bytearray() -- s. :func:`_data_converter`.
            Bytes making up the section.
        start_address: int
        join: bool
            Join/merge adjacent section.

        Raises
        ------
        :class:`InvalidAddressError`

        Notes
        -----
        Overlapping sections are not supported.
        To update/replace a section use :meth:`update_section`.
        """
        start_address = start_address if start_address is not None else self.address  # If Address omitted, create continuous address space.
        if self._address_contained(start_address, len(data)):
            raise InvalidAddressError("Overlapping address-space")
        if isinstance(data, str):
            data = [ord(x) for x in data] # array.array('B',data)
        self._sections.add(Section(start_address, data))
        if join:
            self.join_sections()
        self.address = start_address + len(data)

    @property
    def sections(self):
        return self._sections

    def get_section(self, address):
        """Get :class:`Section` containing `address`.
        Parameters
        ----------
        address: int

        Returns
        -------
        :class:`Section`

        Raises
        ------
        :class:`InvalidAddressError`
        """
        if not address in self:
            raise InvalidAddressError("Address not in range")
        result = self._sections.bisect_right(SearchType(address))
        return self._sections[result - 1]


    def update_section(self, data, address = None):
        """

        """
        if not self._address_contained(address, len(data)):
            raise InvalidAddressError("Address-space not in range")

    def delete_section(self, address = None):
        """

        """

    def join_sections(self):
        """

        """
        self._sections = join_sections(self._sections)

    def split(self, at = None, equal_parts = None, remap = None):
        print("SPLIT-IMAGE", at, equal_parts, remap)

def _validate_sections(sections):
    """Test for required protocol
    """
    ATTRIBUTES = ('start_address', 'length', 'data')
    if not '__iter__' in dir(sections):
        raise TypeError("Sections must be iteratable.")
    for section in sections:
        if not all( hasattr(section, attr) for attr in ATTRIBUTES):
            raise TypeError("Section '{0}' doesn't fulfills required protocol (missing attributes).")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2020 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


from datetime import datetime
from functools import reduce
import os

from objutils.logger import Logger

"""
$00-$7F     Simple number in the range 0 to 127, or 7-bit ASCII string with length 0 to 127.
$80-$84     Number larger than 127 or negative. 0 to 4 bytes follow. $80 is used as a place
            holder and means the value was not provided.
$85-$88     Unsigned numbers between 2ˆˆ32 and (2ˆˆ64)-1, or negative. Not supported in all installations.
$89-$8F     Unused.
$90-$9F     Reserved functions and operators.
$A0-$BF     Function values (arithmetic and logical).
$C0-$DA     Variable letters (null, A-Z).
$DB-$DD     Unused.
$DE-$DF     Extension length. If DE, the next byte is the length of an 8-bit string between 0 and 255
            bytes long. If DF, the next two bytes in high- order/low-order format are the length of
            an 8-bit string between 0 and 65535 bytes long.
$E0-$FA     Record headers.
$FB         Define Context.
$FC-$FF     Unused.

Table 2-1. Initial Bytes of IEEE Elements
"""

SMALL_NUMBER    = 0x7f
SHORT_STRING    = 0x7f
COMMAND_CODE    = 0xe0

"""
ASP     = 0xD0    # Current Section PC.
ASS     = 0xD3    # Section Size.
ASR     = 0xD2    # Variable Values.
ASW     = 0xD7    # Variable Values.
ASA     = 0xC1    # Physical Region Size.
ASB     = 0xC2    # Physical Region Base Address.
ASP     = 0xC6    # Mau Size.
ASG     = 0xC7    # Execution Starting Address.
ASI     = 0xC9    # Variable Values.
ASL     = 0xCC    # Section Base Address.
ASM     = 0xCD    # M-Value.
ASN     = 0xCE
"""

F1      = 0xF1
E2      = 0xE2

##
## Header Part.
##
MB      = 0xE0      # Module Beginning.
AD      = 0xEC      # Address Descriptor.

ASW0    = 0x00    # Assign Pointer to AD Extension Part.
ASW1    = 0x01    # Assign Pointer to Environment Part.
ASW2    = 0x02    # Assign Pointer to Section Part.
ASW3    = 0x03    # Assign Pointer to External Part.
ASW4    = 0x04    # Assign Pointer to Debug Part.
ASW5    = 0x05    # Assign Pointer to Data Part.
ASW6    = 0x06    # Assign Pointer to Trailer Part.
ASW7    = 0x07    # Assign Pointer to Module End.


ASWAssignment = {
    ASW0: "AD Extension",
    ASW1: "Environment",
    ASW2: "Section",
    ASW3: "External",
    ASW4: "Debug",
    ASW5: "Data",
    ASW6: "Trailer",
    ASW7: "Module End",
}

##
## AD Extension Part (ASW0).
##
NN      = 0xF0       # Variable Attributes
ATN     = 0xCE     # Variable Attributes    - F1
ASN     = 0xCE     # Variable Values    - E2


##
## Environment Part (ASW1).
##

## Same as ASW0
"""
NN      = 0xF0      # Variable Attributes
ATN     = 0xF1CE    # Variable Attributes
ASN     = 0xE2CE    # Variable Values
"""

##
## Section Definition Part (ASW2).
##
ST      = 0xE6    # Section Type.
SA      = 0xE7    # Section Alignment.
ASS     = 0xD3    # Section Size.   - E2
ASL     = 0xCC    # Section Base Address.   - E2
ASR     = 0xD2    # Variable Values.  - E2
NC      = 0xFB    # Define Context.
ASA     = 0xC1    # Physical Region Size. - E2
ASB     = 0xC2    # Physical Region Base Address. - E2
ASP     = 0xC6    # Mau Size. - E2
ASF     = 0xC6    # Mau Size. - E2
ASM     = 0xCD    # M-Value.  - E2


##
## External Part (ASW3).
##
NI      = 0xE8    # Public (External) Symbol.
ATI     = 0xC9    # Variable Attribute. - F1
ASI     = 0xC9    # Variable Values.  - E2
NX      = 0xE9    # External Reference Name.
ATX     = 0xD8    # External Reference Relocation Information.  - F1
WX      = 0xF4    # Weak External Reference.


##
## Debug Information Definition Part (ASW4).
##
BB      = 0xF8    # Declare Block Beginning.
NN      = 0xF0    # Declare Type Name, file name, line numbers, function name, variable names, etc.
TY      = 0xF2    # Define Type Characteristics.
ASN     = 0xCE    # Variable Values.  - E2
ASR     = 0xD2    # Variable Values. - E2
BE      = 0xF9    # Declare Block End.


##
## Data Part (ASW5).
##
SB      = 0xE5    # Current Section.
ASP     = 0xD0    # Current Section PC.   - E2
LD      = 0xED    # Load Constant MAUs.
IR      = 0xE3    # Initialize Relocation Base.
RE      = 0xF7    # Repeat Data.
ASW     = 0xD7    # Variable Values.  - E2
LR      = 0xE4    # Load With Relocation.
LT      = 0xFA    # Load With Translation.


##
## Trailer Part (ASW6).
##
ASG     = 0xC7    # Execution Starting Address.   - E2


##
## Module End (ASW7).
##
ME      = 0xE1      # Module End.
##Checksum Records - 0xEE, 0xEF


##
##  Blocktypes.
##
BB1 = 1     # Type definitions local to a module.
BB2 = 2     # Type definitions global to all modules.
BB3 = 3     # A module. A non-separable unit of code, usually the result
            # of a single compilation, i.e. a Modula-2 module or an Ada package.
BB4 = 4     # A global subprogram.
BB5 = 5     # A source file line number block.
BB6 = 6     # A local (static) subprogram.
BB10 = 10   # An assembler debugging information block.
BB11 = 11   # The module portion of a section.
BB20 = 20   # Library - contains a list of global symbols used in a module.


##
##  Object Format Types.
##
ObjectFormatTypes = {
    1: "Absolute (not relinkable)",
    2: "Relocatable",
    3: "Loadable",
    4: "Library"
}

CaseSensitivity = {
    1: "case insensitiv",
    2: "case sensitiv"
}

MemoryModel = {
    0: "tiny",      #Code and data are in the same single 64K segment/page.
    1: "small",     # Code and data each have a single 64K segment/page.
    2: "medium",    # Data has a single 64K segment/page, while code has multiple 64K segments/pages.
    3: "compact",   # Data has multiple 64K segments/pages, while code has a single 64K segment/page.
    4: "large",     # Both data and code have multiple 64K segments/pages.
    5: "big",       # Code has multiple 64K segments/pages, while there is a common "near" data area with
                    # far data areas available; normaly data and stack are together
    6: "huge",      # All large arrays and structures are in their own section so that
                    # addressing involves computations (you can have arrays and structures bigger than 64K)
}

FunctionsAndOperators = {
    0xA0: "@F",
    0xB0: "@AND",
    0xA1: "@T",
    0xB1: "@OR",
    0xA2: "@ABS",
    0xB2: "@XOR",
    0xA3: "@NEG",
    0xB3: "@EXT",
    0xA4: "@NOT",
    0xB4: "@INS",
    0xA5: "+",
    0xB5: "@ERR",
    0xA6: "-",
    0xB6: "@IF",
    0xA7: "/",
    0xB7: "@ELSE",
    0xA8: "*",
    0xB8: "@END",
    0xA9: "@MAX",
    0xB9: "@ESCAPE",
    0xAA: "@MIN",
    0xBA: "[",
    0xAB: "@MOD",
    0xBB: "]",
    0xAC: "<",
    0xBC: "{",
    0xAD: ">",
    0xBD: "}",
    0xAE: "=",
    0xBE: "(",
    0xAF: "!=",
    0xBF: ")",
}


ESCAPEFunctionEncodings = {
    0 : "reserved",
    1 : "@ISDEF",
    2 : "@TRANS",
    3 : "@SPLIT",
    4 : "@INBLOCK",
    5 : "@CALL_OPT",
}


class InvalidFormatError(Exception): pass


class Info(object): pass


class DebugInformation(object):
    def __init__(self):
        self.children = []

    def add(self, child):
        self.children.append(child)


class Section(object):
    def __init__(self, sectionType, sectionName, parentSectionIndex):
        self.sectionType = sectionType
        self.sectionName =  sectionName
        self.parentSectionIndex = parentSectionIndex
        self.regionSize = []
        self.physicalAddress = []
        self.sectionOffset = []
        self.mauSize = 0
        self.mValue = None
        self.segments = {}


class Reader(object):

    def __init__(self, inFile):
        if not hasattr(inFile, 'read'):
            raise TypeError("Need a file-like object.")
        self.logger = Logger("IEEE695")
        self.inFile = inFile
        self.info = Info()
        self.info.ASWs={}
        self.fpos = 0
        self.blockType = []
        self.finished = False
        self.symbols = {}
        self.externalSymbols = {}
        self.sections = {}
        self.nnRecords = {}
        self.contexts = {}
        diRoot = DebugInformation()
        diRoot.parent = None
        diRoot.name = "ROOT"
        self.debugInformation = [diRoot]
        self.diParents = [diRoot]
        self.currentSection = None
        self.currentSectionIndex = None
    ####
        self._nb = 0
        self.dbCollection = bytearray()

        while not self.finished:
            cc = self.readByte(self.fpos)
            if cc == MB:
                self.onModuleBegin()
            elif cc == AD:
                self.onAD()
            elif cc == E2:
                self.onE2()
            elif cc == NX:
                self.onNX()
            elif cc == WX:
                self.onWX()
            elif cc == NN:
                self.onNN()
            elif cc == F1:
                self.onF1()
            elif cc == ST:
                self.onST()
            elif cc == NC:
                self.onNC()
            elif cc == SA:
                self.onSA()
            elif cc == NI:
                self.onNI()
            elif cc == BB:
                self.onBB()
            elif cc == BE:
                self.onBE()
            elif cc == TY:
                self.onTY()
            elif cc == SB:
                self.onSB()
            elif cc == LD:
                self.onLD()
            elif cc == ME:
                self.onME()
            else:
                raise NotImplementedError("0x{0:02X}".format(cc))


    def setCurrentSectionIndex(self, index):
        self.currentSectionIndex = index

    def checkSectionIndex(self, index):
        if  self.currentSectionIndex is None:
            print("No current Section Index.")
        elif self.currentSectionIndex != index:
            print
            print("Invalid Section Index.")


    def readByte(self, offset):
        "Read 8bit quantity"
        self.inFile.seek(offset, os.SEEK_SET)
        result = ord(self.inFile.read(1))
        self.fpos = self.inFile.tell()
        return result

    def readWord(self, offset):
        "Read 16bit quantity"
        self.inFile.seek(offset, os.SEEK_SET)
        h = ord(self.inFile.read(1))
        l = ord(self.inFile.read(1))
        self.fpos = self.inFile.tell()
        return h << 8 | l

    def readNumber(self, offset):
        "Read number of abitrary length"
        self.inFile.seek(offset, os.SEEK_SET)
        typecode = self.readByte(offset)
        if typecode <= SMALL_NUMBER:
            # length: [0..127]
            result = typecode
        elif 0x80 <= typecode <= 0x88:
            # length (in bytes) [0..8]
            length = typecode & ~0x80
            data = self.inFile.read(length)
            self.fpos = self.inFile.tell()
            result = reduce(lambda x, y: (x * 256) + ord(y), data, 0)
        return result

    def readString(self, offset):
        typecode = self.readByte(offset)

        if typecode <= SHORT_STRING:
            # length: [0..127]
            length = typecode
        elif typecode == 0xde:
            # length: [0..255]
            length = self.readByte(offset + 1)
        elif typecode == 0xdf:
            # length: [0..65535]
            length = self.readWord(offset + 1)
        else:
            raise TypeError("Invalid typecode [{0:02x}].".format(typecode))
        result = self.inFile.read(length)
        self.fpos = self.inFile.tell()
        return result

    def readCharacter(self, offset):
        result = chr(self.readByte(offset) - 0x80)
        if ord(result) > ord('Z'):
            result = None
        return result

    def checkOptional(self, offset, isString = False):
        "Read and maybe putback"
        cc = self.readByte(offset)
        if cc >= COMMAND_CODE:
            self.inFile.seek(offset, os.SEEK_SET)   # Put back.
            self.fpos = offset
            return None
        if isString == True:
            self.inFile.seek(offset, os.SEEK_SET)   # Put back.
            self.fpos = offset
            return self.readString(offset)
        else:
            if 0xC0 <= cc <= 0xDA:
                result = chr(cc - 0x80)
                if ord(result) > ord('Z'):
                    result = None
            else:
                return self.readNumber(offset)

    def symbolByName(self, name):
        pass

    def symbolByIndex(self, index):
        pass

################################
################################
################################
    def onModuleBegin(self):
        "{$E0}{Id1}{Id2}"
        self.info.processor = self.readString(self.fpos)
        self.info.module = self.readString(self.fpos)

        self.logger.debug("PROCESSOR: '{0!s}' MODULE: '{1!s}'.".format(self.info.processor, self.info.module))

    def onAD(self):
        "$EC}{n1}{n2}[a]"
        self.info.numberOfBits = self.readNumber(self.fpos)
        self.info.numberOfMAUs = self.readNumber(self.fpos)
        self.info.adressWidth = self.info.numberOfBits * self.info.numberOfMAUs
        self.info.byteOrder = self.checkOptional(self.fpos)

    def onE2(self):
        "{$E2}{$D7}{0x}{n}"
        """
        ASR     = 0xD2    # Variable Values.
        """
        discr = self.readByte(self.fpos)
        if discr == ASW:
            addr = self.readByte(self.fpos)
            byteOffset = self.readNumber(self.fpos)
            self.info.ASWs[addr] = (ASWAssignment[addr], byteOffset)
        elif discr == ASS:
            sectionIndex = self.readNumber(self.fpos)
            self.checkSectionIndex(sectionIndex)
            self.sections[sectionIndex].sectionSize = self.readNumber(self.fpos)
            self.logger.debug("Section-Size: 0x{0:04x}".format(self.sections[sectionIndex].sectionSize))
        elif discr == ASA:
            sectionIndex = self.readNumber(self.fpos)
            self.checkSectionIndex(sectionIndex)
            self.sections[sectionIndex].regionSize.append(self.readNumber(self.fpos))
        elif discr == ASB:
            sectionIndex = self.readNumber(self.fpos)
            self.checkSectionIndex(sectionIndex)
            self.sections[sectionIndex].physicalAddress.append(self.readNumber(self.fpos))
        elif discr == ASR:
            sectionIndex = self.readNumber(self.fpos)
            self.checkSectionIndex(sectionIndex)
            self.sections[sectionIndex].sectionOffset.append(self.readNumber(self.fpos))
        elif discr == ASF:
            sectionIndex = self.readNumber(self.fpos)
            self.checkSectionIndex(sectionIndex)
            self.sections[sectionIndex].mauSize = self.readNumber(self.fpos)
            self.logger.debug("MAU-Size: 0x{0:04x}".format(self.sections[sectionIndex].mauSize))
        elif discr == ASL:
            sectionIndex = self.readNumber(self.fpos)
            self.checkSectionIndex(sectionIndex)
            self.sections[sectionIndex].sectionBaseAddr = self.readNumber(self.fpos)
            self.logger.debug("Section-BaseAddr: 0x{0:04x}".format(self.sections[sectionIndex].sectionBaseAddr))
        elif discr == ASM:
            sectionIndex = self.readNumber(self.fpos)
            self.checkSectionIndex(sectionIndex)
            self.sections[sectionIndex].mValue = self.readNumber(self.fpos)
        elif discr == ASI:
            # There may be more than one ASI/ATI record pair for a single
            # name since constants defined in SETs and #defines my be redefined repeatedly.
            symbolIndex = self.readNumber(self.fpos)
            expr = self.readNumber(self.fpos)
            self.symbols[symbolIndex].expr = expr
        elif discr == ASN:
            symbolNameIndex = self.readNumber(self.fpos)
            symbolTypeIndex = self.readNumber(self.fpos)
        elif discr == ASP:
            sectionIndex = self.readNumber(self.fpos)   #? check sec-Index!!?
            newValue = self.readNumber(self.fpos)
        elif discr == ASG:
            delim = self.readByte(self.fpos)
            if delim != 0xBE:
                pass    # todo: FormatError!!!
            executionStartingAddr = self.readNumber(self.fpos)
            self.logger.debug("STARTING-ADDRESS: 0x{0:04X}".format(executionStartingAddr))
            delim = self.readByte(self.fpos)
            if delim != 0xBF:
                pass    # todo: FormatError!!!
            self.info.executionStartingAddr = executionStartingAddr
        else:
            raise NotImplementedError(hex(discr))

    def onNN(self):
        "{$F0}{n1}{Id}"
        symbolType = self.readNumber(self.fpos)
        symbolName = self.readString(self.fpos)

        self.nnRecords[symbolType] = Info()
        self.nnRecords[symbolType].name = symbolName
        self.nnRecords[symbolType].atn = {}

        #self.logger.debug("Symbol-Type: 0x%02x Name: '%s'" % (symbolType, symbolName))

    def onNX(self):
        "{$E9}{n1}{Id}"
        externalReferenceIndex = self.readNumber(self.fpos)
        symbolName = self.readString(self.fpos)
        self.externalSymbols[externalReferenceIndex] = Info()
        self.externalSymbols[externalReferenceIndex].symbolName = symbolName

    def onWX(self):
        "{$F4}{n1}{n2}{n3}"
        externalReferenceIndex = self.readNumber(self.fpos)
        defaultSize = self.readNumber(self.fpos)
        defaultValue = self.readNumber(self.fpos)
        self.externalSymbols[externalReferenceIndex].defaultSize = defaultSize
        self.externalSymbols[externalReferenceIndex].defaultValue = defaultValue

    def onF1(self):
        """
            ATI     = 0xF1C9    # Variable Attribute.
            ATX     = 0xF1D8    # External Reference Relocation Information.
            ATN     = 0xF1CE    # Variable Attributes.
        """
        discr = self.readByte(self.fpos)
        if discr == ATI:
            numberOfElements = 0
            symbolNameIndex = self.readNumber(self.fpos)
            symbolTypeIndex =self.readNumber(self.fpos)
            attrDef =self.readNumber(self.fpos)
            if attrDef == 8:
                pass
            elif attrDef == 16:
                symbolClass = self.readNumber(self.fpos)
                publicLocalIndicator =  self.checkOptional(self.fpos)
                numericValue = self.checkOptional(self.fpos)
                stringValue = self.checkOptional(self.fpos)

                self.symbols[symbolNameIndex].symbolClass = symbolClass
                self.symbols[symbolNameIndex].publicLocalIndicator = publicLocalIndicator
                self.symbols[symbolNameIndex].numericValue = numericValue
                self.symbols[symbolNameIndex].stringValue = stringValue
            elif attrDef == 19:
                pass
            else:
                pass    # todo: FormatError!!!
##            n4 =self.readNumber(self.fpos) # If n2 is non-zero, number of elements in the symbol type specified in n2
            if symbolTypeIndex != 0:
                numberOfElements = self.readNumber(self.fpos)
            self.symbols[symbolNameIndex].symbolTypeIndex = symbolTypeIndex
            self.symbols[symbolNameIndex].numberOfElements = numberOfElements
            self.symbols[symbolNameIndex].attrDef = attrDef
        elif discr == ATX:
            externalReferenceIndex = self.readNumber(self.fpos)
            typeIndex = self.readNumber(self.fpos)
            sectionIndex = self.readNumber(self.fpos)
            shortExternalFlag = self.readNumber(self.fpos)
            self.externalSymbols[externalReferenceIndex].typeIndex = typeIndex
            self.externalSymbols[externalReferenceIndex].sectionIndex = sectionIndex
            self.externalSymbols[externalReferenceIndex].shortExternalFlag = shortExternalFlag
        elif discr == ATN:  # todo: onATN
            symbolNameIndex = self.readNumber(self.fpos)
            symbolTypeIndex =self.readNumber(self.fpos)
            attrDef =self.readNumber(self.fpos)
            if attrDef == 1:
                stackOffset = self.readNumber(self.fpos)
            elif attrDef == 2:
                registerIndex = self.readNumber(self.fpos)
            elif attrDef == 3:
                pass
#            elif attrDef == 4:
#                pass
            elif attrDef == 7:
                lineNumber = self.readNumber(self.fpos)
                columnNumber = self.readNumber(self.fpos)
                res1 = self.checkOptional(self.fpos)
                res2 = self.checkOptional(self.fpos)
            elif attrDef == 8:
                pass
            elif attrDef == 9:
                absoluteProgramCounterOffset = self.readNumber(self.fpos)
                ##if symbolNameIndex == 0:
                ##    registerResourceIndex = self.readByte(self.fpos)
            elif attrDef == 10:
                registerIndex = self.readNumber(self.fpos)
                frameOffset = self.readNumber(self.fpos)
#            elif attrDef == 11:
#                pass
            elif attrDef == 19:
                numberOfElements = self.readNumber(self.fpos)
                localGlobal = self.checkOptional(self.fpos)
            elif attrDef == 37:
                self.info.objectFormatVersionNumber = self.readNumber(self.fpos)
                self.info.objectFormatRevisionLevel  = self.readNumber(self.fpos)
            elif attrDef == 38:
                self.info.objectFormatType=ObjectFormatTypes[self.readNumber(self.fpos)]
            elif attrDef == 39:
                self.info.symbolCaseSensitivity = CaseSensitivity[self.readNumber(self.fpos)]
            elif attrDef == 40:
                self.info.memoryModel = MemoryModel[self.readNumber(self.fpos)]
            elif attrDef == 50:
                year = self.readNumber(self.fpos)
                month = self.readNumber(self.fpos)
                day = self.readNumber(self.fpos)
                hour = self.readNumber(self.fpos)
                minute = self.readNumber(self.fpos)
                second = self.readNumber(self.fpos)
                self.info.creationDate = datetime(year, month, day, hour, minute, second)
            elif attrDef == 51:
                self.info.commandLine = self.readString(self.fpos)
            elif attrDef == 52:
                self.info.executionStatus = self.readNumber(self.fpos)
            elif attrDef == 53:
                self.info.hostEnvironment = self.readNumber(self.fpos)
            elif attrDef == 54:
                tool = self.readNumber(self.fpos)
                version = self.readNumber(self.fpos)
                revision = self.readNumber(self.fpos)
                revisionLevel = self.checkOptional(self.fpos)   # single IEEE-695 letter in the range $C1-$DA
            elif attrDef == 55:
                self.info.comments = self.readString(self.fpos)
            elif attrDef == 64:
                typeIdent = self.readNumber(self.fpos)
                addATN = self.readNumber(self.fpos)
            elif attrDef == 65:
                miscString = self.readString(self.fpos)
            else:
                raise NotImplementedError("Invalid ATN-Attr: 0x{0:02x}".format(attrDef))
                # todo: FormatError
        else:
            raise NotImplementedError(hex(discr))
            # todo: FormatError

    def onST(self):
        "${E6}{n1}{l}[Id][n2][n3][n4]"
        sectionIndex = self.readByte(self.fpos)
        self.setCurrentSectionIndex(sectionIndex)
        f = self.readCharacter(self.fpos)
        if f == 'A':
            s, t = self.readCharacter(self.fpos), self.readCharacter(self.fpos)
            sectionType = f + s + t
        elif f == 'B':
            pass
        elif f == 'C':
            pass
        elif f == 'E':
            pass # todo: 'EA' / 'EZ' !!!
        elif f == 'M':
            pass
        elif f == 'T':
            pass
        #todo: 'ZC' / 'ZM'.
        else:
            raise NotImplementedError("SEG-TYPE: {0!s}".format(f))

        sectionName = self.readString(self.fpos)
        parentSectionIndex = self.checkOptional(self.fpos)

        if f == 'T':
            brotherSectionIndex = self.readNumber(self.fpos)

        if f in ('B', 'T'):
            contextIndex = self.readNumber(self.fpos)

        self.sections[sectionIndex] = Section(sectionType, sectionName, parentSectionIndex)
        self.logger.debug("SECTION [{0!s}:{1!s}]".format(sectionType, sectionName))
        # SA, ASA, ASB, ASF, ASL, ASM, ASR, and ASS records must appear after the ST record they refer to.
        '''
        ASP absolute code
        ASR absolute ROM data
        ASD absolute data

        CP code
        CR ROM data
        CD data

        EAP common absolute code
        EAR common absolute ROM data
        EAD common absolute data

        ZCP  short code
        ZCR  short ROM data
        ZCD  short data
        '''

    def onNC(self):
        "{$FB}{n1}{Id}"
        contextIndex = self.readNumber(self.fpos)
        contextName = self.readString(self.fpos)
        self.contexts[contextIndex] = contextName

    def onSA(self):
        "{$E7}{n1}[n2][n3]"
        sectionIndex = self.readByte(self.fpos)
        self.checkSectionIndex(sectionIndex)
        boundaryAlignmentDivisor = self.checkOptional(self.fpos)
        self.sections[sectionIndex].boundaryAlignment = 2 ** boundaryAlignmentDivisor
        self.sections[sectionIndex].pageSize  = self.checkOptional(self.fpos)

    def onNI(self):
        "{$E8}{n}{Id}"
        nameIndex = self.readNumber(self.fpos)  # must be > 31, 0 - 31 reserved.
        symbolName = self.readString(self.fpos)
        info = Info()
        info.nameIndex = nameIndex
        info.symbolName = symbolName
        self.symbols[nameIndex] = info
        self.logger.debug("SYMBOL: {0!s}".format(symbolName)) # followed by ASI.

    def onBB(self):
        blockType = self.readByte(self.fpos)
        blockSize = self.readNumber(self.fpos)
        info = DebugInformation()
        info.parent = self.diParents[-1]
        info.blockType = blockType
        info.blockSize = blockSize
        if blockSize == 0:
            pass
        if blockType == BB1:
            self.blockType.append(1)
            module_name = self.readString(self.fpos)
            info.module_name = module_name
            info.name = "BB1"
            self.logger.debug("MODULE-NAME: {0!s}".format(module_name))
        elif blockType == BB2:
            self.blockType.append(2)
            zeroLengthName = self.readString(self.fpos)
            info.zeroLengthName = zeroLengthName
            info.name = "BB2"
        elif blockType == BB3:
            self.blockType.append(3)
            module_name = self.readString(self.fpos)
            info.module_name = module_name
            info.name = "BB3"
        elif blockType == BB4:
            self.blockType.append(4)
            functionName = self.readString(self.fpos)
            stackSpaceRequired = self.readNumber(self.fpos)
            typeIndexOfReturnValue = self.readNumber(self.fpos)
            offsetExpr = self.readNumber(self.fpos)
            info.functionName = functionName
            info.stackSpaceRequired = stackSpaceRequired
            info.typeIndexOfReturnValue = typeIndexOfReturnValue
            info.offsetExpr = offsetExpr
            info.name = "BB4"
        elif blockType ==  BB5:
            self.blockType.append(5)
            sourceFilename = self.readString(self.fpos)
            year = self.checkOptional(self.fpos)    # todo: factor-out (namedtuple)!!!
            month = self.checkOptional(self.fpos)
            day = self.checkOptional(self.fpos)
            hour = self.checkOptional(self.fpos)
            minute = self.checkOptional(self.fpos)
            second = self.checkOptional(self.fpos)
            info.sourceFilename = sourceFilename
            info.year = year
            info.month = month
            info.day = day
            info.hour = hour
            info.minute = minute
            info.second = second
            info.name = "BB5"
        elif blockType ==  BB6:
            self.blockType.append(6)
            functionName = self.readString(self.fpos)
            stackRequired = self.readNumber(self.fpos)
            typeIndexForReturnValue = self.readNumber(self.fpos)
            offsetExpr = self.readNumber(self.fpos)
            info.functionName = functionName
            info.stackRequired = stackRequired
            info.typeIndexForReturnValue = typeIndexForReturnValue
            info.offsetExpr = offsetExpr
            info.name = "BB6"
        elif blockType == BB10:
            self.blockType.append(10)
            module_name = self.readString(self.fpos)
            objFileName = self.readString(self.fpos)
            toolType = self.readNumber(self.fpos)
            versionRevision = self.checkOptional(self.fpos, isString = True)
            year = self.checkOptional(self.fpos)
            month = self.checkOptional(self.fpos)
            day = self.checkOptional(self.fpos)
            hour = self.checkOptional(self.fpos)
            minute = self.checkOptional(self.fpos)
            second = self.checkOptional(self.fpos)
            info.module_name = module_name
            info.objFileName = objFileName
            info.toolType = toolType
            info.versionRevision = versionRevision
            info.year = year
            info.month = month
            info.day = day
            info.hour = hour
            info.minute = minute
            info.second = second
            info.name = "BB10"
        elif blockType ==  BB11:
            self.blockType.append(11)
            zeroLengthName = self.readString(self.fpos)
            sectionType = self.readNumber(self.fpos)
            sectionIndex = self.readNumber(self.fpos)
            offsetExpr = self.readNumber(self.fpos)
            dummy = self.readByte(self.fpos)   # $90 - Comma operator!!!
            sectionMapping = self.readNumber(self.fpos)
            info.zeroLengthName = zeroLengthName
            info.sectionType = sectionType
            info.sectionIndex = sectionIndex
            info.offsetExpr = offsetExpr
            info.dummy = dummy
            info.sectionMapping = sectionMapping
            info.name = "BB11"
        elif blockType ==  BB20:
            self.blockType.append(20)
            raise NotImplementedError('BB20')
        parent = self.diParents[-1]
        parent.add(info)
        self.diParents.append(info)
        #self.logger.debug(" " * len(self.diParents), "BB{0:d}".format(blockType))

    def onBE(self):
        blockType = self.blockType.pop()
        self.diParents.pop()
        if  blockType in (4, 6):
            functionEndAddr = self.readNumber(self.fpos)
        elif blockType == 11:
            moduleSectionSize = self.checkOptional(self.fpos)

    def onTY(self):
        ""
        typeIndex = self.readNumber(self.fpos)
        if self.readByte(self.fpos) != 0xCE:
            pass    # todo: raise FormatError!!
        localNameIndex = self.readNumber(self.fpos)
        values = []
        while True:
            value = self.checkOptional(self.fpos)
            if value is None:
                break
            else:
                values.append(value)

    def onSB(self):
        "{$E5}{n1}"
        sectionIndex = self.readNumber(self.fpos)
        sec = self.sections[sectionIndex]
        self.logger.debug("Data for section: '{0!s}'. {1:d} bytes of data to follow.".format(
            sec.sectionName, sec.sectionSize
        ))

    def onLD(self):
        "{$ED}{n1}{...}"
        numberOfMAUs = self.readNumber(self.fpos)
        data = self.inFile.read((numberOfMAUs * self.info.numberOfBits) / 8)
        self.logger.debug("reading {0:d} bytes".format(len(data)))


        # SB ASP LD
        self._nb += len(data)
        self.dbCollection.extend(data)

        self.fpos = self.inFile.tell()


    def onME(self):
        "Module End Record Type"
        self.finished = True
        self.logger.debug("{0:d} Data-Bytes.".format(self._nb))


## $C0-$DA Variable letters (null, A-Z).

## Information-Variables: s. 3

"""
Number Format:
--------------
•   If the value is between 0-127 decimal, the number is $0-$7F.
•   If the value is greater than 127 decimal, then the number must be defined by 1 byte of count with the high
    order bit set ($80) followed by the indicated number of bytes of numeric data with the most significant byte
    first. The range for the count is usually 0-4 (i.e. $80-$84) and can be 0-8 on some installations. This form is
    also valid for numbers in the range 0-127.
    Example: $7FFF is encoded as {$82}{$7F}{$FF} (3 bytes). 0 can be encoded as {$00} or {$81}{$00}, 2ˆˆ32
    can be encoded as {$85}{01}{00}{00}{00}{00}, etc.
•   Omitted optional fields in records may be represented by a byte count of zero.
    Example: {$80}

Negative numbers have these characteristics:
    • Numbers use $80-$88.
    • All readers must handle $80-$84.
    • Some processors can use $85 through $88 (see the appendix for your processor).
    • Numbers are construed as unsigned except where indicated as signed.
    • Signed numbers use signed complement notation; negative numbers must set highest bit of largest
      representable number (e.g., {$84}{$FF}{$FF}{$FF}{$FF} is -1 for 32-bit and smaller processors).

"""




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from objutils import loads, dumps, probes
import unittest

import pytest

#for key, value in sorted(registry.registry().items()):
#    print("{0:10} {1}".format(key, value.description))


SREC = b"""S123B000576F77212044696420796F75207265616C6C7920676F207468726F7567682061DF
S120B0206C6C20746861742074726F75626C6520746F207265616420746869733F73
S5030002FA"""

FPC = b"""$kL&@h%%,:,B.\?00EPuX0K3rO0JI))
$;UPR'%%,:<Hn&FCG:at<GVF(;G9wIw
$7FD1p%%,:LHmy:>GTV%/KJ7@GE[kYz
$B[6\;%%,:\KIn?GFWY/qKI1G5:;-_e
$%%%%%"""


fromSrec =  loads('srec', SREC)
dataFromSRec = dumps('fpc', fromSrec)   # Fixme: Does _NOT_ return 'bytes'!

fromFPC =  loads('fpc', FPC)
dataFromFPC = dumps('srec', fromFPC, row_length = 32, s5record = True)


class TestRoundTrip(unittest.TestCase):

    def testFromSrec(self):
        self.assertEqual(dataFromSRec, FPC)
        #pass

    def testFromFPC(self):
        self.assertEqual(dataFromFPC, SREC)


class TestProbe(unittest.TestCase):

    @pytest.mark.skip
    def testProbeSrec(self):
        self.assertEqual(probes(SREC), "srec")

    @pytest.mark.skip
    def testProbeFpc(self):
        self.assertEqual(probes(FPC), "fpc")


if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

##
##  ASCII Space Hex format.
##

import objutils.hexfile as hexfile


class Reader(hexfile.ASCIIHexReader):
    """
    """

    def __init__(self, address_pattern = r'^@(?P<address>[0-9a-zA-Z]{2,8})\s*$',
                 data_pattern = r'^(?:[0-9a-zA-Z]{{2,4}}[{0}]?)*\s*$', etx_pattern = r'^q.*$'):
        super(Reader, self).__init__(address_pattern, data_pattern, etx_pattern)


class Writer(hexfile.ASCIIHexWriter):

    def __init__(self, address_designator = '@'):
        super(Writer, self).__init__(address_designator)

    def compose_footer(self, meta):
        return "q"




import unittest

class DummyReader(object): pass
class DummyWriter(object): pass

class TestRegistry(unittest.TestCase):
    CODECS = ('A', 'B', 'C', 'X', 'Y', 'Z')

    """
    def tearDown(self):
        self.clearRegistry()

    def testRegistryShallBeEmpty(self):
        self.assertEqual(objutils.registry.registry(), {})

    def registerSomeCodecs(self):
        for cdc in self.CODECS:
            objutils.registry.register(cdc, DummyReader, DummyWriter)

    def clearRegistry(self):
        objutils.registry.codecs = {}

    def testRegistration(self):
        self.registerSomeCodecs()
        self.assertEqual(sorted(objutils.registry.getFormats()), ['A', 'B', 'C', 'X', 'Y', 'Z'])

    def testGetCodec(self):
        self.registerSomeCodecs()
        self.assertIsInstance(objutils.registry.getCodec('C'), objutils.registry.Codec)

    def testCodecNotFound(self):
        self.registerSomeCodecs()
        with self.assertRaises(objutils.registry.CodecDoesNotExistError):
            cdc =  objutils.registry.getCodec('Foo')

    def testCodecAlreadyExists(self):
        self.registerSomeCodecs()
        with self.assertRaises(objutils.registry.CodecAlreadyExistError):
            objutils.registry.register('X', DummyReader, DummyWriter)
    """

if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import enum
from collections import namedtuple
import struct


ELF_MAGIC = b'\x7fELF'

EI_NIDENT = 16        # Size of e_ident[].
HDR_FMT32 = "HHIIIIIHHHHHH"
HDR_FMT64 = "HHIQQQIHHHHHH"


Elf32_Ehdr = namedtuple("Elf32_Ehdr", """e_type e_machine e_version e_entry e_phoff e_shoff e_flags e_ehsize
                        e_phentsize e_phnum e_shentsize e_shnum e_shstrndx""")

class ELFType(enum.IntEnum):
    ET_NONE     = 0
    " No file type"
    ET_REL      = 1
    " Relocatable file"
    ET_EXEC     = 2
    " Executable file"
    ET_DYN      = 3
    " Shared object file"
    ET_CORE     = 4
    " Core file"
    ET_NUM      = 5
    "Number of defined types"
    ET_LOOS     = 0xFE00
    "Operating system-specific"
    ET_HIOS     = 0xFEFF
    "Operating system-specific"
    ET_LOPROC   = 0xff00
    " Processor-specific"
    ET_HIPROC   = 0xffff
    "Processor-specific"


ELF_TYPE_NAMES = {
    ELFType.ET_NONE : "No file type",
    ELFType.ET_REL  : "Relocatable file",
    ELFType.ET_EXEC : "Executable file",
    ELFType.ET_DYN  : "Shared object file",
    ELFType.ET_CORE : "Core file"
}


class ELFMachineType(enum.IntEnum):
    EM_NONE         =  0      # No machine.
    EM_M32          =  1      # AT&T WE 32100.
    EM_SPARC        =  2      # SPARC.
    EM_386          =  3      # Intel 80386.
    EM_68K          =  4      # Motorola 68000.
    EM_88K          =  5      # Motorola 88000.

    '''
    RESERVED 6 Reserved for future use
    '''

    EM_860          =  7      # Intel 80860.
    EM_MIPS         =  8      # MIPS I Architecture.
    EM_S370         =  9      # IBM System/370 Processor.
    EM_MIPS_RS3_LE  =  10     # MIPS RS3000 Little-endian.

    '''
    RESERVED 11-14 Reserved for future use
    '''

    EM_PARISC       =  15     # Hewlett-Packard PA-RISC.
    RESERVED        =  16     # Reserved for future use.
    EM_VPP500       =  17     # Fujitsu VPP500.
    EM_SPARC32PLUS  =  18     # Enhanced instruction set SPARC.
    EM_960          =  19     # Intel 80960.
    EM_PPC          =  20     # PowerPC.
    EM_PPC64        =  21     # 64-bit PowerPC.
    EM_S390         =  22     # IBM S390.
    EM_SPU          =  23     # Sony/Toshiba/IBM SPU.

    '''
    RESERVED 24-35 Reserved for future use
    '''

    EM_V800         =  36     # NEC V800.
    EM_FR20         =  37     # Fujitsu FR20.
    EM_RH32         =  38     # TRW RH-32.
    EM_RCE          =  39     # Motorola RCE.
    EM_ARM          =  40     # Advanced RISC Machines ARM.
    EM_ALPHA        =  41     # Digital Alpha.
    EM_SH           =  42     # Hitachi SH.
    EM_SPARCV9      =  43     # SPARC Version 9.
    EM_TRICORE      =  44     # Siemens Tricore embedded processor.
    EM_ARC          =  45     # Argonaut RISC Core, Argonaut Technologies Inc.
    EM_H8_300       =  46     # Hitachi H8/300.
    EM_H8_300H      =  47     # Hitachi H8/300H.
    EM_H8S          =  48     # Hitachi H8S.
    EM_H8_500       =  49     # Hitachi H8/500.
    EM_IA_64        =  50     # Intel IA-64 processor architecture.
    EM_MIPS_X       =  51     # Stanford MIPS-X.
    EM_COLDFIRE     =  52     # Motorola ColdFire.
    EM_68HC12       =  53     # Motorola M68HC12.    # could also be 0x4D12 (s. HC12EABI)
    EM_MMA          =  54     # Fujitsu MMA Multimedia Accelerator.
    EM_PCP          =  55     # Siemens PCP.
    EM_NCPU         =  56     # Sony nCPU embedded RISC processor.
    EM_NDR1         =  57     # Denso NDR1 microprocessor.
    EM_STARCORE     =  58     # Motorola Star*Core processor.
    EM_ME16         =  59     # Toyota ME16 processor.
    EM_ST100        =  60     # STMicroelectronics ST100 processor.
    EM_TINYJ        =  61     # Advanced Logic Corp. TinyJ embedded processor family.
    EM_X86_64       =  62     # Advanced Micro Devices X86-64.
    EM_PDSP         =  63     # Sony DSP Processor.
    EM_PDP10        =  64     # DEC PDP-10
    EM_PDP11        =  65     # DEC PDP-11
    EM_FX66         =  66     # Siemens FX66 microcontroller.
    EM_ST9PLUS      =  67     # STMicroelectronics ST9+ 8/16 bit microcontroller.
    EM_ST7          =  68     # STMicroelectronics ST7 8-bit microcontroller.
    EM_68HC16       =  69     # Motorola MC68HC16 Microcontroller.
    EM_68HC11       =  70     # Motorola MC68HC11 Microcontroller.
    EM_68HC08       =  71     # Motorola MC68HC08 Microcontroller.
    EM_68HC05       =  72     # Motorola MC68HC05 Microcontroller.
    EM_SVX          =  73     # Silicon Graphics SVx.
    EM_ST19         =  74     # STMicroelectronics ST19 8-bit microcontroller.
    EM_VAX          =  75     # Digital VAX.
    EM_CRIS         =  76     # Axis Communications 32-bit embedded processor.
    EM_JAVELIN      =  77     # Infineon Technologies 32-bit embedded processor.
    EM_FIREPATH     =  78     # Element 14 64-bit DSP Processor.
    EM_ZSP          =  79     # LSI Logic 16-bit DSP Processor.
    EM_MMIX         =  80     # Donald Knuth's educational 64-bit processor.
    EM_HUANY        =  81     # Harvard University machine-independent object files .
    EM_PRISM        =  82     # SiTera Prism.
    EM_AVR          =  83     # Atmel AVR 8-bit microcontroller.
    EM_FR30         =  84     # Fujitsu FR30.
    EM_D10V         =  85     # Mitsubishi D10V.
    EM_D30V         =  86     # Mitsubishi D30V.
    EM_V850         =  87     # NEC v850.
    EM_M32R         =  88     # Mitsubishi M32R.
    EM_MN10300      =  89     # Matsushita MN10300.
    EM_MN10200      =  90     # Matsushita MN10200.
    EM_PJ           =  91     # picoJava.
    EM_OPENRISC     =  92     # OpenRISC 32-bit embedded processor.
    EM_ARC_A5       =  93     # ARC Cores Tangent-A5.
    EM_XTENSA       =  94     # Tensilica Xtensa Architecture.
    EM_VIDEOCORE    =  95     # Alphamosaic VideoCore processor.
    EM_TMM_GPP      =  96     # Thompson Multimedia General Purpose Processor.
    EM_NS32K        =  97     # National Semiconductor 32000 series.
    EM_TPC          =  98     # Tenor Network TPC processor.
    EM_SNP1K        =  99     # Trebia SNP 1000 processor.
    EM_ST200        = 100     # STMicroelectronics ST200 microcontroller.
    EM_IP2K         = 101     # Ubicom IP2022 micro controller.
    EM_MAX          = 102     # MAX Processor.
    EM_CR           = 103     # National Semiconductor CompactRISC.
    EM_F2MC16       = 104     # Fujitsu F2MC16.
    EM_MSP430       = 105     # TI msp430 micro controller.
    EM_BLACKFIN     = 106     # ADI Blackfin.
    EM_SE_C33       = 107     # S1C33 Family of Seiko Epson processors.
    EM_SEP          = 108     # Sharp embedded microprocessor.
    EM_ARCA         = 109     # Arca RISC Microprocessor.
    EM_UNICORE      = 110     # Microprocessor series from PKU-Unity Ltd. and MPRC of Peking University.
    EM_EXCESS       = 111     # eXcess: 16/32/64-bit configurable embedded CPU.
    EM_DXP          = 112     # Icera Semiconductor Inc. Deep Execution Processor.
    EM_ALTERA_NIOS2 = 113     # Altera Nios II soft-core processor.
    EM_CRX          = 114     # National Semiconductor CRX.
    EM_XGATE        = 115     # Motorola XGATE embedded processor.
    EM_C166         = 116     # Infineon C16x/XC16x processor.
    EM_M16C         = 117     # Renesas M16C series microprocessors.
    EM_DSPIC30F     = 118     # Microchip Technology dsPIC30F Digital Signal Controller.
    EM_CE           = 119     # Freescale Communication Engine RISC core.
    EM_M32C         = 120     # Renesas M32C series microprocessors.
    EM_TSK3000      = 131     # Altium TSK3000 core.
    EM_RS08         = 132     # Freescale RS08 embedded processor.
    EM_ECOG2        = 134     # Cyan Technology eCOG2 microprocessor.
    EM_SCORE        = 135     # Sunplus Score.
    EM_SCORE7       = 135     # Sunplus S+core7 RISC processor.
    EM_DSP24        = 136     # New Japan Radio (NJR) 24-bit DSP Processor.
    EM_VIDEOCORE3   = 137     # Broadcom VideoCore III processor.
    EM_LATTICEMICO32= 138     # RISC processor for Lattice FPGA architecture.
    EM_SE_C17       = 139     # Seiko Epson C17 family.
    EM_TI_C6000     = 140     # Texas Instruments TMS320C6000 DSP family.
    EM_TI_C2000     = 141     # Texas Instruments TMS320C2000 DSP family.
    EM_TI_C5500     = 142     # Texas Instruments TMS320C55x DSP family.
    EM_MMDSP_PLUS   = 160     # STMicroelectronics 64bit VLIW Data Signal Processor.
    EM_CYPRESS_M8C  = 161     # Cypress M8C microprocessor.
    EM_R32C         = 162     # Renesas R32C series microprocessors.
    EM_TRIMEDIA     = 163     # NXP Semiconductors TriMedia architecture family.
    EM_QDSP6        = 164     # QUALCOMM DSP6 Processor.
    EM_I8051        = 165     # Intel 8051 and variants.
    EM_STXP7X       = 166     # STMicroelectronics STxP7x family.
    EM_NDS32        = 167     # Andes Technology compact code size embedded RISC processor family.
    EM_ECOG1        = 168     # Cyan Technology eCOG1X family.
    EM_ECOG1X       = 168     # Cyan Technology eCOG1X family.
    EM_MAXQ30       = 169     # Dallas Semiconductor MAXQ30 Core Micro-controllers.
    EM_XIMO16       = 170     # New Japan Radio (NJR) 16-bit DSP Processor.
    EM_MANIK        = 171     # M2000 Reconfigurable RISC Microprocessor.
    EM_CRAYNV2      = 172     # Cray Inc. NV2 vector architecture.
    EM_RX           = 173     # Renesas RX family.
    EM_METAG        = 174     # Imagination Technologies META processor architecture.
    EM_MCST_ELBRUS  = 175     # MCST Elbrus general purpose hardware architecture.
    EM_ECOG16       = 176     # Cyan Technology eCOG16 family.
    EM_CR16         = 177     # National Semiconductor CompactRISC 16-bit processor.
    EM_ETPU         = 178     # Freescale Extended Time Processing Unit.
    EM_SLE9X        = 179     # Infineon Technologies SLE9X core.
    EM_L1OM         = 180     # Intel L1OM.
    EM_AARCH64      = 183
    EM_AVR32        = 185     # Atmel Corporation 32-bit microprocessor family.
    EM_STM8         = 186     # STMicroeletronics STM8 8-bit microcontroller.
    EM_TILE64       = 187     # Tilera TILE64 multicore architecture family.
    EM_TILEPRO      = 188     # Tilera TILEPro multicore architecture family.
    EM_MICROBLAZE   = 189     # Xilinx MicroBlaze 32-bit RISC soft processor core.
    EM_CUDA         = 190     # NVIDIA CUDA architecture.
    EM_TILEGX       = 191     # Tilera TILE-Gx
    EM_AVR_OLD              = 0x1057  # AVR magic number.  Written in the absense of an ABI.
    EM_MSP430_OLD           = 0x1059  # MSP430 magic number.  Written in the absense of everything.
    EM_MT                   = 0x2530  # Morpho MT.   Written in the absense of an ABI.
    EM_CYGNUS_FR30          = 0x3330  # FR30 magic number - no EABI available.
    EM_OPENRISC_OLD         = 0x3426  # OpenRISC magic number.  Written in the absense of an ABI.
    EM_DLX                  = 0x5aa5  # DLX magic number.  Written in the absense of an ABI.
    EM_CYGNUS_FRV           = 0x5441  # FRV magic number - no EABI available??.
    EM_XC16X                = 0x4688  # Infineon Technologies 16-bit microcontroller with C166-V2 core.
    EM_NEC_V850             = 0x723e  # V850ES/FJ3 /w NEC compiler.
    EM_CYGNUS_D10V          = 0x7650  # D10V backend magic number.  Written in the absence of an ABI.
    EM_CYGNUS_D30V          = 0x7676  # D30V backend magic number.  Written in the absence of an ABI.
    EM_IP2K_OLD             = 0x8217  # Ubicom IP2xxx;   Written in the absense of an ABI.
    EM_OR32                 = 0x8472  # (Deprecated) Temporary number for the OpenRISC processor.
    EM_CYGNUS_POWERPC       = 0x9025  # Cygnus PowerPC ELF backend.  Written in the absence of an ABI.
    EM_ALPHA_NO_ABI         = 0x9026  # Alpha backend magic number.  Written in the absence of an ABI.
    EM_CYGNUS_M32R          = 0x9041  # Cygnus M32R ELF backend.  Written in the absence of an ABI.
    EM_CYGNUS_V850          = 0x9080  # V850 backend magic number.  Written in the absense of an ABI.
    EM_S390_OLD             = 0xa390  # old S/390 backend magic number. Written in the absence of an ABI.
    EM_XTENSA_OLD           = 0xabc7  # Old, unofficial value for Xtensa.
    EM_XSTORMY16            = 0xad45
    EM_CYGNUS_MN10300       = 0xbeef  # mn10200 and mn10300 backend magic numbers. Written in the absense of an ABI.
    EM_COOLRISC_C816        = 0xc816  # Raisonance CoolRISC C816.
    EM_CYGNUS_MN10200       = 0xdead
    EM_M32C_OLD             = 0xFEB0  # Renesas M32C and M16C.
    EM_IQ2000               = 0xFEBA  # Vitesse IQ2000.
    EM_NIOS32               = 0xFEBB  # NIOS magic number - no EABI available.
    EM_CYGNUS_MEP           = 0xF00D  # Toshiba MeP
    EM_MOXIE                = 0xFEED  # Moxie
    EM_MICROBLAZE_OLD       = 0xbaab  # Old MicroBlaze
    EM_ADAPTEVA_EPIPHANY    = 0x1223  # Adapteva's Epiphany architecture.


ELF_MACHINE_NAMES = {
    ELFMachineType.EM_NONE          : "No machine",
    ELFMachineType.EM_M32           : "AT&T WE 32100",
    ELFMachineType.EM_SPARC         : "SPARC",
    ELFMachineType.EM_386           : "Intel 80386",
    ELFMachineType.EM_68K           : "Motorola 68000",
    ELFMachineType.EM_88K           : "Motorola 88000",
    ELFMachineType.EM_860           : "Intel 80860",
    ELFMachineType.EM_MIPS          : "MIPS I Architecture",
    ELFMachineType.EM_S370          : "IBM System/370 Processor",
    ELFMachineType.EM_MIPS_RS3_LE   : "MIPS RS3000 Little-endian",
    ELFMachineType.EM_PARISC        : "Hewlett-Packard PA-RISC",
    ELFMachineType.RESERVED         : "Reserved for future use",
    ELFMachineType.EM_VPP500        : "Fujitsu VPP500",
    ELFMachineType.EM_SPARC32PLUS   : "Enhanced instruction set SPARC",
    ELFMachineType.EM_960           : "Intel 80960",
    ELFMachineType.EM_PPC           : "PowerPC",
    ELFMachineType.EM_PPC64         : "PowerPC64",
    ELFMachineType.EM_S390          : "IBM S390",
    ELFMachineType.EM_SPU           : "Sony/Toshiba/IBM SPU",
    ELFMachineType.EM_V800          : "Renesas V850 (using RH850 ABI)",
    ELFMachineType.EM_FR20          : "Fujitsu FR20",
    ELFMachineType.EM_RH32          : "TRW RH-32",
    ELFMachineType.EM_RCE           : "Motorola RCE",
    ELFMachineType.EM_ARM           : "ARM",
    ELFMachineType.EM_ALPHA         : "Digital Alpha",
    ELFMachineType.EM_SH            : "Hitachi SH",
    ELFMachineType.EM_SPARCV9       : "SPARC Version 9",
    ELFMachineType.EM_TRICORE       : "Siemens Tricore embedded processor",
    ELFMachineType.EM_ARC           : "Argonaut RISC Core, Argonaut Technologies Inc",
    ELFMachineType.EM_H8_300        : "Hitachi H8/300",
    ELFMachineType.EM_H8_300H       : "Hitachi H8/300H",
    ELFMachineType.EM_H8S           : "Hitachi H8S",
    ELFMachineType.EM_H8_500        : "Hitachi H8/500",
    ELFMachineType.EM_IA_64         : "Intel IA-64 processor architecture",
    ELFMachineType.EM_MIPS_X        : "Stanford MIPS-X",
    ELFMachineType.EM_COLDFIRE      : "Motorola ColdFire",
    ELFMachineType.EM_68HC12        : "Motorola MC68HC12 Microcontroller",
    ELFMachineType.EM_MMA           : "Fujitsu MMA Multimedia Accelerator",
    ELFMachineType.EM_PCP           : "Siemens PCP",
    ELFMachineType.EM_NCPU          : "Sony nCPU embedded RISC processor",
    ELFMachineType.EM_NDR1          : "Denso NDR1 microprocessor",
    ELFMachineType.EM_STARCORE      : "Motorola Star*Core processor",
    ELFMachineType.EM_ME16          : "Toyota ME16 processor",
    ELFMachineType.EM_ST100         : "STMicroelectronics ST100 processor",
    ELFMachineType.EM_TINYJ         : "Advanced Logic Corp. TinyJ embedded processor family",
    ELFMachineType.EM_X86_64        : "Advanced Micro Devices X86-64",
    ELFMachineType.EM_PDSP          : "Sony DSP Processor",
    ELFMachineType.EM_PDP10         : "DEC PDP-10",
    ELFMachineType.EM_PDP11         : "DEC PDP-11",
    ELFMachineType.EM_FX66          : "Siemens FX66 microcontroller",
    ELFMachineType.EM_ST9PLUS       : "STMicroelectronics ST9+ 8/16 bit microcontroller",
    ELFMachineType.EM_ST7           : "STMicroelectronics ST7 8-bit microcontroller",
    ELFMachineType.EM_68HC16        : "Motorola MC68HC16 Microcontroller",
    ELFMachineType.EM_68HC11        : "Motorola MC68HC11 Microcontroller",
    ELFMachineType.EM_68HC08        : "Motorola MC68HC08 Microcontroller",
    ELFMachineType.EM_68HC05        : "Motorola MC68HC05 Microcontroller",
    ELFMachineType.EM_SVX           : "Silicon Graphics SVx",
    ELFMachineType.EM_ST19          : "STMicroelectronics ST19 8-bit microcontroller",
    ELFMachineType.EM_VAX           : "Digital VAX",
    ELFMachineType.EM_CRIS          : "Axis Communications 32-bit embedded processor",
    ELFMachineType.EM_JAVELIN       : "Infineon Technologies 32-bit embedded processor",
    ELFMachineType.EM_FIREPATH      : "Element 14 64-bit DSP Processor",
    ELFMachineType.EM_ZSP           : "LSI Logic 16-bit DSP Processor",
    ELFMachineType.EM_MMIX          : "Donald Knuth's educational 64-bit processor",
    ELFMachineType.EM_HUANY         : "Harvard University machine-independent object files",
    ELFMachineType.EM_PRISM         : "SiTera Prism",
    ELFMachineType.EM_AVR           : 'Atmel AVR 8-bit microcontroller',
    ELFMachineType.EM_FR30          : 'Fujitsu FR30',
    ELFMachineType.EM_D10V          : 'Mitsubishi D10V',
    ELFMachineType.EM_D30V          : 'Mitsubishi D30V',
    ELFMachineType.EM_V850          : 'NEC v850',
    ELFMachineType.EM_M32R          : 'Mitsubishi M32R',
    ELFMachineType.EM_MN10300       : 'Matsushita MN10300',
    ELFMachineType.EM_MN10200       : 'Matsushita MN10200',
    ELFMachineType.EM_PJ            : 'picoJava',
    ELFMachineType.EM_OPENRISC      : 'OpenRISC 32-bit embedded processor',
    ELFMachineType.EM_ARC_A5        : 'ARC Cores Tangent-A5',
    ELFMachineType.EM_XTENSA        : 'Tensilica Xtensa Architecture',
    ELFMachineType.EM_VIDEOCORE     : 'Alphamosaic VideoCore processor',
    ELFMachineType.EM_TMM_GPP       : 'Thompson Multimedia General Purpose Processor',
    ELFMachineType.EM_NS32K         : 'National Semiconductor 32000 series',
    ELFMachineType.EM_TPC           : 'Tenor Network TPC processor',
    ELFMachineType.EM_SNP1K         : 'Trebia SNP 1000 processor',
    ELFMachineType.EM_ST200         : 'STMicroelectronics ST200 microcontroller',
    ELFMachineType.EM_IP2K          : 'Ubicom IP2022 micro controller',
    ELFMachineType.EM_MAX           : 'MAX Processor',
    ELFMachineType.EM_CR            : 'National Semiconductor CompactRISC',
    ELFMachineType.EM_F2MC16        : 'Fujitsu F2MC16',
    ELFMachineType.EM_MSP430        : 'TI msp430 micro controller',
    ELFMachineType.EM_BLACKFIN      : 'ADI Blackfin',
    ELFMachineType.EM_SE_C33        : 'S1C33 Family of Seiko Epson processors',
    ELFMachineType.EM_SEP           : 'Sharp embedded microprocessor',
    ELFMachineType.EM_ARCA          : 'Arca RISC Microprocessor',
    ELFMachineType.EM_UNICORE       : 'Microprocessor series from PKU-Unity Ltd. and MPRC of Peking University',
    ELFMachineType.EM_EXCESS        : 'eXcess: 16/32/64-bit configurable embedded CPU',
    ELFMachineType.EM_DXP           : 'Icera Semiconductor Inc. Deep Execution Processor',
    ELFMachineType.EM_ALTERA_NIOS2  : 'Altera Nios II soft-core processor',
    ELFMachineType.EM_CRX           : 'National Semiconductor CRX',
    ELFMachineType.EM_XGATE         : 'Motorola XGATE embedded processor',
    ELFMachineType.EM_C166          : 'Infineon Technologies xc16x',
    ELFMachineType.EM_M16C          : 'Renesas M16C series microprocessors',
    ELFMachineType.EM_DSPIC30F      : 'Microchip Technology dsPIC30F Digital Signal Controller',
    ELFMachineType.EM_CE            : 'Freescale Communication Engine RISC core',
    ELFMachineType.EM_M32C          : 'Renesas M32C series microprocessors',
    ELFMachineType.EM_TSK3000       : 'Altium TSK3000 core',
    ELFMachineType.EM_RS08          : 'Freescale RS08 embedded processor',
    ELFMachineType.EM_ECOG2         : 'Cyan Technology eCOG2 microprocessor',
    ELFMachineType.EM_SCORE         : 'Sunplus Score',
    ELFMachineType.EM_SCORE7        : 'Sunplus S+core7 RISC processor',
    ELFMachineType.EM_DSP24         : 'New Japan Radio (NJR) 24-bit DSP Processor',
    ELFMachineType.EM_VIDEOCORE3    : 'Broadcom VideoCore III processor',
    ELFMachineType.EM_LATTICEMICO32 : 'RISC processor for Lattice FPGA architecture',
    ELFMachineType.EM_SE_C17        : 'Seiko Epson C17 family',
    ELFMachineType.EM_TI_C6000      : 'Texas Instruments TMS320C6000 DSP family',
    ELFMachineType.EM_TI_C2000      : 'Texas Instruments TMS320C2000 DSP family',
    ELFMachineType.EM_TI_C5500      : 'Texas Instruments TMS320C55x DSP family',
    ELFMachineType.EM_MMDSP_PLUS    : 'STMicroelectronics 64bit VLIW Data Signal Processor',
    ELFMachineType.EM_CYPRESS_M8C   : 'Cypress M8C microprocessor',
    ELFMachineType.EM_R32C          : 'Renesas R32C series microprocessors',
    ELFMachineType.EM_TRIMEDIA      : 'NXP Semiconductors TriMedia architecture family',
    ELFMachineType.EM_QDSP6         : 'QUALCOMM DSP6 Processor',
    ELFMachineType.EM_I8051         : 'Intel 8051 and variants',
    ELFMachineType.EM_STXP7X        : 'STMicroelectronics STxP7x family',
    ELFMachineType.EM_NDS32         : 'Andes Technology compact code size embedded RISC processor family',
    ELFMachineType.EM_ECOG1         : 'Cyan Technology eCOG1X family',
    ELFMachineType.EM_ECOG1X        : 'Cyan Technology eCOG1X family',
    ELFMachineType.EM_MAXQ30        : 'Dallas Semiconductor MAXQ30 Core Micro-controllers',
    ELFMachineType.EM_XIMO16        : 'New Japan Radio (NJR) 16-bit DSP Processor',
    ELFMachineType.EM_MANIK         : 'M2000 Reconfigurable RISC Microprocessor',
    ELFMachineType.EM_CRAYNV2       : 'Cray Inc. NV2 vector architecture',
    ELFMachineType.EM_RX            : 'Renesas RX family',
    ELFMachineType.EM_METAG         : 'Imagination Technologies META processor architecture',
    ELFMachineType.EM_MCST_ELBRUS   : 'MCST Elbrus general purpose hardware architecture',
    ELFMachineType.EM_ECOG16        : 'Cyan Technology eCOG16 family',
    ELFMachineType.EM_CR16          : 'National Semiconductor CompactRISC 16-bit processor',
    ELFMachineType.EM_ETPU          : 'Freescale Extended Time Processing Unit',
    ELFMachineType.EM_SLE9X         : 'Infineon Technologies SLE9X core',
    ELFMachineType.EM_L1OM          : 'Intel L1OM',
    ELFMachineType.EM_AARCH64       : 'AArch64',
    ELFMachineType.EM_AVR32         : 'Atmel Corporation 32-bit microprocessor family',
    ELFMachineType.EM_STM8          : 'STMicroeletronics STM8 8-bit microcontroller',
    ELFMachineType.EM_TILE64        : 'Tilera TILE64 multicore architecture family',
    ELFMachineType.EM_TILEPRO       : 'Tilera TILEPro multicore architecture family',
    ELFMachineType.EM_MICROBLAZE    : 'Xilinx MicroBlaze 32-bit RISC soft processor core',
    ELFMachineType.EM_TILEGX        : 'Tilera TILE-Gx',
    ELFMachineType.EM_CUDA          : 'NVIDIA CUDA architecture',
    ELFMachineType.EM_AVR_OLD       : 'AVR',
    ELFMachineType.EM_MSP430_OLD    : 'MSP430',
    ELFMachineType.EM_MT            : 'Morpho MT',
    ELFMachineType.EM_CYGNUS_FR30   : 'Cygnus FR30',
    ELFMachineType.EM_OPENRISC_OLD  : 'OpenRISC',
    ELFMachineType.EM_DLX           : 'DLX',
    ELFMachineType.EM_CYGNUS_FRV    : 'Cygnus FRV',
    ELFMachineType.EM_XC16X         : 'Infineon C166-V2 core',
    ELFMachineType.EM_NEC_V850      : 'V850ES/FJ3 (NEC compiler)',
    ELFMachineType.EM_CYGNUS_D10V   : 'Cygnus D10V',
    ELFMachineType.EM_CYGNUS_D30V   : 'Cygnus D30V',
    ELFMachineType.EM_IP2K_OLD      : 'Ubicom IP2xxx',
    ELFMachineType.EM_OR32          : 'OpenRISC 32',
    ELFMachineType.EM_CYGNUS_POWERPC: 'Cygnus PowerPC',
    ELFMachineType.EM_ALPHA_NO_ABI  : 'Alpha',
    ELFMachineType.EM_CYGNUS_M32R   : 'Cygnus M32R',
    ELFMachineType.EM_CYGNUS_V850   : 'Cygnus V850',
    ELFMachineType.EM_S390_OLD      : 'S/390',
    ELFMachineType.EM_XTENSA_OLD    : 'Xtensa',
    ELFMachineType.EM_XSTORMY16     : 'Xstormy16',
    ELFMachineType.EM_CYGNUS_MN10300: 'Cygnus mn10200 or mn10300',
    ELFMachineType.EM_COOLRISC_C816 : 'CoolRISC C816',
    ELFMachineType.EM_CYGNUS_MN10200: 'Cygnus mn10200',
    ELFMachineType.EM_M32C_OLD      : 'Renesas M32C or M16C.',
    ELFMachineType.EM_IQ2000        : 'Vitesse IQ2000.',
    ELFMachineType.EM_NIOS32        : 'NIOS',
    ELFMachineType.EM_CYGNUS_MEP    : 'Toshiba MeP',
    ELFMachineType.EM_MOXIE         : 'Moxie',
    ELFMachineType.EM_MICROBLAZE_OLD: 'Old MicroBlaze',
    ELFMachineType.EM_ADAPTEVA_EPIPHANY : "Adapteva Epiphany",
}


EV_NONE         = 0      # Invalid version.
EV_CURRENT      = 1      # Current version.


##
## Offsets into file header.
##
EI_MAG0         = 0      # File identification.
EI_MAG1         = 1      # File identification.
EI_MAG2         = 2      # File identification.
EI_MAG3         = 3      # File identification.
EI_CLASS        = 4      # File class.
EI_DATA         = 5      # Data encoding.
EI_VERSION      = 6      # File version.
EI_PAD          = 7      # Start of padding bytes.
EI_OSABI        = 7      # Operating system/ABI identification.
EI_ABIVERSION   = 8      # ABI version.
# EI_NIDENT       = 16     # Size of e_ident[] - defined above.


class ELFClass(enum.IntEnum):
    ELFCLASSNONE    = 0      # Invalid class.
    ELFCLASS32      = 1      # 32-bit objects.
    ELFCLASS64      = 2      # 64-bit objects.


ELF_CLASS_NAMES = {
    ELFClass.ELFCLASSNONE   : "Invalid class",
    ELFClass.ELFCLASS32     : "32-bit objects",
    ELFClass.ELFCLASS64     : "64-bit objects"
}


class ELFDataEncoding(enum.IntEnum):
    ELFDATANONE     = 0      # Invalid data encoding.
    ELFDATA2LSB     = 1      # Little-Endian.
    ELFDATA2MSB     = 2      # Big-Endian.


ELF_BYTE_ORDER_NAMES = {
    ELFDataEncoding.ELFDATANONE : "Invalid data encoding",
    ELFDataEncoding.ELFDATA2LSB : "Little-Endian",
    ELFDataEncoding.ELFDATA2MSB : "Big-Endian"
}

class ELFAbiType(enum.IntEnum):
    ELFOSABI_NONE         = 0   # UNIX System V ABI
    ELFOSABI_HPUX         = 1   # HP-UX operating system
    ELFOSABI_NETBSD       = 2   # NetBSD
    ELFOSABI_GNU          = 3   # GNU
    ELFOSABI_LINUX        = 3   # Alias for ELFOSABI_GNU
    ELFOSABI_SOLARIS      = 6   # Solaris
    ELFOSABI_AIX          = 7   # AIX
    ELFOSABI_IRIX         = 8   # IRIX
    ELFOSABI_FREEBSD      = 9   # FreeBSD
    ELFOSABI_TRU64        = 10  # TRU64 UNIX
    ELFOSABI_MODESTO      = 11  # Novell Modesto
    ELFOSABI_OPENBSD      = 12  # OpenBSD
    ELFOSABI_OPENVMS      = 13  # OpenVMS
    ELFOSABI_NSK          = 14  # Hewlett-Packard Non-Stop Kernel
    ELFOSABI_AROS         = 15  # AROS
    ELFOSABI_FENIXOS      = 16  # FenixOS
    ELFOSABI_C6000_ELFABI = 64  # Bare-metal TMS320C6000; alt:  ELFOSABI_ARM_AEABI
    ELFOSABI_C6000_LINUX  = 65  # Linux TMS320C6000
    ELFOSABI_ARM          = 97  # ARM
    ELFOSABI_STANDALONE   = 255 # Standalone (embedded) application

##
##
##   ELF Sections.
##
##

SEC_FMT32 = "IIIIIIIIII"
ELF_SECTION_SIZE32 = struct.calcsize(SEC_FMT32)

SEC_FMT64 = "IIQQQQIIQQ"
ELF_SECTION_SIZE64 = struct.calcsize(SEC_FMT64)

Elf_Shdr = namedtuple("Elf_Shdr", """sh_name sh_type sh_flags sh_addr sh_offset sh_size
    sh_link sh_info sh_addralign sh_entsize"""
)


class SectionName(enum.IntEnum):
    SHN_UNDEF       = 0
        # SHN_UNDEF This value marks an undefined, missing, irrelevant, or otherwise
        # meaningless section reference. For example, a symbol "defined'' relative to
        # section number SHN_UNDEF is an undefined symbol.

    SHN_LORESERVE   = 0xff00
        # This value specifies the lower bound of the range of reserved indexes
    SHN_LOPROC      = 0xff00
    SHN_BEFORE      = 0xff00
        # Order section before all others (Solaris).
    SHN_AFTER       = 0xff01
        # Order section after all others (Solaris).
    SHN_HIPROC      = 0xff1f
        # SHN_LOPROC through SHN_HIPROC: Values in this inclusive range are reserved for
        # processor-specific semantics.
    SHN_LOOS        = 0xff20
    # Start of OS-specific
    SHN_HIOS        = 0xff3f
    # End of OS-specific
    SHN_ABS         = 0xfff1
        # SHN_ABS This value specifies absolute values for the corresponding reference. For
        # example, symbols defined relative to section number SHN_ABS have
        # absolute values and are not affected by relocation.
    SHN_COMMON      = 0xfff2
        # SHN_COMMON Symbols defined relative to this section are common symbols, such as
        # FORTRAN COMMON or unallocated C external variables.
    SHN_XINDEX      = 0xffff
        # Index is in extra table.
    SHN_HIRESERVE   = 0xffff
        # SHN_HIRESERVE This value specifies the upper bound of the range of reserved indexes.
        # The system reserves indexes between SHN_LORESERVE and SHN_HIRESERVE, inclusive;
        # the values do not reference the section header table.That is, the section header
        # table does not contain entries for the  reserved indexes.

    SHN_IA_64_ANSI_COMMON   = 0xFF00
        # This section only used by HP-UX, The HP linker gives weak symbols
        # precedence over regular common symbols.  We want common to override
        # weak.  Using this common instead of SHN_COMMON does that.

SpecialSections = {v: k for k, v in SectionName.__members__.items()}

def special_section_name(ndx):
    if ndx in SpecialSections:
        return SectionName(ndx).name
    else:
        return str(ndx)

class SectionType(enum.IntEnum):
    """
    """
    SHT_NULL            = 0             # Section header table entry unused.
    SHT_PROGBITS        = 1             # Program data.
    SHT_SYMTAB          = 2             # Symbol table.
    SHT_STRTAB          = 3             # String table.
    SHT_RELA            = 4             # Relocation entries with addends.
    SHT_HASH            = 5             # Symbol hash table.
    SHT_DYNAMIC         = 6             # Dynamic linking information.
    SHT_NOTE            = 7             # Notes.
    SHT_NOBITS          = 8             # Program space with no data (bss).
    SHT_REL             = 9             # Relocation entries, no addends.
    SHT_SHLIB           = 10            # Reserved.
    SHT_DYNSYM          = 11            # Dynamic linker symbol table.
    SHT_INIT_ARRAY      = 14            # Array of constructors.
    SHT_FINI_ARRAY      = 15            # Array of destructors.
    SHT_PREINIT_ARRAY   = 16            # Array of pre-constructors.
    SHT_GROUP           = 17            # Section group.
    SHT_SYMTAB_SHNDX    = 18            # Extended section indeces.
    SHT_NUM             = 19            # Number of defined types.
    SHT_LOOS            = 0x60000000    # Start OS-specific.
    SHT_GNU_INCREMENTAL_INPUTS = 0x6fff4700 # incremental build data
    SHT_GNU_ATTRIBUTES  = 0x6ffffff5    # Object attributes.
    SHT_GNU_HASH        = 0x6ffffff6    # GNU-style hash table.
    SHT_GNU_LIBLIST     = 0x6ffffff7    # Prelink library list
    SHT_CHECKSUM        = 0x6ffffff8    # Checksum for DSO content.
    SHT_LOSUNW          = 0x6ffffffa    # Sun-specific low bound.
    SHT_SUNW_move       = 0x6ffffffa
    SHT_SUNW_COMDAT     = 0x6ffffffb
    SHT_SUNW_syminfo    = 0x6ffffffc
    SHT_GNU_verdef      = 0x6ffffffd    # Version definition section.
    SHT_GNU_verneed     = 0x6ffffffe    # Version needs section.
    SHT_GNU_versym      = 0x6fffffff    # Version symbol table.
    SHT_HISUNW          = 0x6fffffff    # Sun-specific high bound.
    SHT_HIOS            = 0x6fffffff    # End OS-specific type.
    SHT_LOPROC          = 0x70000000    # Start of processor-specific.

    SHT_ARM_EXIDX       = 0x70000001    # Section holds ARM unwind info.
    SHT_ARM_PREEMPTMAP  = 0x70000002    # Section pre-emption details.
    SHT_ARM_ATTRIBUTES  = 0x70000003    # Section holds attributes.
    SHT_ARM_DEBUGOVERLAY    = 0x70000004    # Section holds overlay debug info.
    SHT_ARM_OVERLAYSECTION  = 0x70000005    # Section holds GDB and overlay integration info.

    SHT_HIPROC          = 0x7fffffff    # End of processor-specific.
    SHT_LOUSER          = 0x80000000    # Start of application-specific.
    SHT_HIUSER          = 0xffffffff    # End of application-specific.


class SectionFlags(enum.IntEnum):
    SHF_WRITE               = 0x1           # Writeable.
    SHF_ALLOC               = 0x2           # Occupies memory during execution
    SHF_EXECINSTR           = 0x4           # Executable.

    SHF_MERGE               = 16            # Might be merged
    SHF_STRINGS             = 32            # Contains nul-terminated strings
    SHF_INFO_LINK           = 64            # `sh_info' contains SHT index
    SHF_LINK_ORDER          = 128           # Preserve order after combining
    SHF_OS_NONCONFORMING    = 256           # Non-standard OS specific handling required
    SHF_GROUP               = 512           # Section is member of a group.
    SHF_TLS                 = 1024          # Section hold thread-local data.
    SHF_COMPRESSED          = 2048          # Section with compressed data.

    SHF_MASKOS              = 0x0ff00000    # OS-specific.
    SHF_MASKPROC            = 0xf0000000    # Processor-specific.

    SHF_ORDERED             = 1073741824    # Special ordering requirement (Solaris).
    SHF_EXCLUDE             = 2147483648    # Section is excluded unless referenced or allocated (Solaris).

##
##
##   ELF Symbol Table.
##
##

SYMTAB_FMT32 = "IIIBBH"
SYMTAB_FMT64 = "IBBHQQ"

Elf32_Sym = namedtuple("Elf_Sym", "st_name st_value st_size st_info st_other st_shndx")
Elf64_Sym = namedtuple("Elf_Sym", "st_name st_info st_other st_shndx st_value st_size")

ELF32_SYM_TABLE_SIZE = struct.calcsize(SYMTAB_FMT32)
ELF64_SYM_TABLE_SIZE = struct.calcsize(SYMTAB_FMT64)

STN_UNDEF           = 0

class SymbolBinding(enum.IntEnum):
    STB_LOCAL           = 0     # i.e. `static`
    STB_GLOBAL          = 1
    STB_WEAK            = 2
    STB_GNU_UNIQUE      = 10    # Symbol is unique in namespace.

    STB_LOPROC          = 13
    STB_HIPROC          = 15

class SymbolType(enum.IntEnum):
    STT_NOTYPE          = 0
    STT_OBJECT          = 1
    STT_FUNC            = 2
    STT_SECTION         = 3
    STT_FILE            = 4
    STT_COMMON          = 5
    STT_TLS             = 6
    STT_RELC            = 8     # Complex relocation expression.
    STT_SRELC           = 9     # Signed Complex relocation expression.
    STT_GNU_IFUNC       = 10    # Symbol is an indirect code object.
    STT_LOPROC          = 13
    STT_HIPROC          = 15

class SymbolVisibility(enum.IntEnum):
    """Symbol visibility specification encoded in the st_other field.
    """
    STV_DEFAULT         = 0 # Default symbol visibility rules
    STV_INTERNAL        = 1 # Processor specific hidden class
    STV_HIDDEN          = 2 # Sym unavailable in other modules
    STV_PROTECTED       = 3 # Not preemptible, not exported

##
##
##   ELF Program Header
##
##
PT_NULL             = 0             # Program header table entry unused.
PT_LOAD             = 1             # Loadable program segment.
PT_DYNAMIC          = 2             # Dynamic linking information.
PT_INTERP           = 3             # Program interpreter.
PT_NOTE             = 4             # Auxiliary information.
PT_SHLIB            = 5             # Reserved.
PT_PHDR             = 6             # Entry for header table itself.
PT_TLS              = 7             # Thread-local storage segment
PT_NUM              = 8             # Number of defined types
PT_LOOS             = 0x60000000    # Start of OS-specific
PT_GNU_EH_FRAME     = 0x6474e550    # GCC .eh_frame_hdr segment
PT_GNU_STACK        = 0x6474e551    # Indicates stack executability
PT_GNU_RELRO        = 0x6474e552    # Read-only after relocation
PT_GNU_PROPERTY     = 0x6474e553    # GNU property

PT_LOSUNW           = 0x6ffffffa
PT_SUNWBSS          = 0x6ffffffa    # Sun Specific segment
PT_SUNWSTACK        = 0x6ffffffb    # Stack segment
PT_HISUNW           = 0x6fffffff
PT_HIOS             = 0x6fffffff    # End of OS-specific
PT_LOPROC           = 0x70000000    # Start of processor-specific.
PT_ARM_EXIDX        = PT_LOPROC + 1 # Frame unwind information
PT_HIPROC           = 0x7fffffff    # End of processor-specific


PF_X                = 0x1           # Execute.
PF_W                = 0x2           # Write.
PF_R                = 0x4           # Read.
PF_MASKPROC         = 0xf0000000    # Unspecified.

PN_XNUM             = 0xffff        # Extended numbering.


class NoteSegmentDescriptorCore(enum.IntEnum):

    """
    Legal values for note segment descriptor types for core files.
    """
    NT_PRSTATUS     = 1             # prstatus struct.
    NT_FPREGSET     = 2             # fpregset struct.
    NT_PRPSINFO     = 3             # prpsinfo struct.
    NT_PRXREG       = 4             # prxregset struct.
    NT_TASKSTRUCT   = 4             # task structure.
    NT_PLATFORM     = 5             # String from sysinfo(SI_PLATFORM).
    NT_AUXV         = 6             # auxv array.
    NT_GWINDOWS     = 7             # gwindows struct.
    NT_ASRS         = 8             # asrset struct.
    NT_PSTATUS      = 10            # pstatus struct.
    NT_PSINFO       = 13            # psinfo struct.
    NT_PRCRED       = 14            # prcred struct.
    NT_UTSNAME      = 15            # utsname struct.
    NT_LWPSTATUS    = 16            # lwpstatus struct.
    NT_LWPSINFO     = 17            # lwpinfo struct.
    NT_PRFPXREG     = 20            # fprxregset struct.
    NT_PRXFPREG     = 0x46e62b7f    # user_fxsr_struct.
    NT_PPC_VMX      = 0x100         # PowerPC Altivec/VMX registers.
    NT_PPC_SPE      = 0x101         # PowerPC SPE/EVR registers.
    NT_PPC_VSX      = 0x102         # PowerPC VSX registers.
    NT_386_TLS      = 0x200         # i386 TLS slots (struct user_desc).
    NT_386_IOPERM   = 0x201         # x86 io permission bitmap (1=deny).
    NT_X86_XSTATE   = 0x202         # x86 extended state using xsave.


class NoteSegmentDescriptorObject(enum.IntEnum):

    """
    Legal values for the note segment descriptor types for object files.
    """
    NT_VERSION      = 1             # Contains a version string.


BYTEORDER_PREFIX = {
    ELFDataEncoding.ELFDATA2LSB : '<',  # Little-Endian.
    ELFDataEncoding.ELFDATA2MSB : '>'   # Big-Endian.
}




#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

__version__ = "0.1.1"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import math
import sys


isprintable = lambda ch: 0x1F < ch < 0x80

def unpack(*args):
    return args

class Dumper(object):

    def __init__(self, fp = sys.stdout, num_address_bits = 32):
        self._fp = fp
        self._rolloverMask = 2 ** num_address_bits
        self._nibbles = num_address_bits >> 2
        self._addressMask = "%0{0:d}x ".format(self._nibbles)
        self.previous_row = bytes()  # bytearray()
        self.elided = False

    def dump_data(self, section, offset = 0):
        end = section.length
        line_count = math.ceil(len(section.data) / self.LINE_LENGTH)
        start_pos = 0
        line_num = 0
        end_pos = self.LINE_LENGTH
        while end_pos < end:
            line_num += 1
            row = section.data[start_pos : end_pos]
            if row == self.previous_row:
                if not self.elided:
                    print("          *", file = self._fp)
                    self.elided = True
            else:
                self.dump_row(row, start_pos + section.start_address)
                self.elided = False
            start_pos = end_pos
            end_pos = end_pos + self.LINE_LENGTH
            self.previous_row = row
        row = section.data[start_pos : end_pos]
        self.dump_row(row, start_pos + section.start_address)
        print("-" * 15, file = self._fp)
        print("{0:-9d} bytes".format(section.length), file = self._fp)
        print("-" * 15, file = self._fp)


class CanonicalDumper(Dumper):
    LINE_LENGTH = 0x10

    def printhex_bytes(self, row):
        row = list(row)
        filler = list([0x20] * (self.LINE_LENGTH - len(row)))
        print('|{0}|'.format(('%s' * self.LINE_LENGTH) % unpack(*[isprintable(x) and chr(x) or '.' for x in row + filler] )), file = self._fp)

    def dump_row(self, row, startAddr):
        start_pos = 0
        print(self._addressMask % ((start_pos + startAddr) % self._rolloverMask), file = self._fp, end = " "),
        print('%02x ' * len(row) % unpack(*row), file = self._fp, end = " "),
        if len(row) == 0:
            print("", file = self._fp, end = "")
        if len(row) < self.LINE_LENGTH:
            spaces = "   " * (self.LINE_LENGTH - len(row))
            print(spaces, file = self._fp, end = ""),
        self.printhex_bytes(row)
